from .constraints.constraint import Constraint


def greater_than(version1, version2):
    """
    Evaluates the expression: version1 > version2.

    :type version1: str
    :type version2: str

    :rtype: bool
    """
    return compare(version1, '>', version2)


def greater_than_or_equal(version1, version2):
    """
    Evaluates the expression: version1 >= version2.

    :type version1: str
    :type version2: str

    :rtype: bool
    """
    return compare(version1, '>=', version2)


def less_than(version1, version2):
    """
    Evaluates the expression: version1 < version2.

    :type version1: str
    :type version2: str

    :rtype: bool
    """
    return compare(version1, '<', version2)


def less_than_or_equal(version1, version2):
    """
    Evaluates the expression: version1 <= version2.

    :type version1: str
    :type version2: str

    :rtype: bool
    """
    return compare(version1, '<=', version2)


def equal(version1, version2):
    """
    Evaluates the expression: version1 == version2.

    :type version1: str
    :type version2: str

    :rtype: bool
    """
    return compare(version1, '==', version2)


def not_equal(version1, version2):
    """
    Evaluates the expression: version1 != version2.

    :type version1: str
    :type version2: str

    :rtype: bool
    """
    return compare(version1, '!=', version2)


def compare(version1, operator, version2):
    """
    Evaluates the expression: $version1 $operator $version2

    :type version1: str
    :type operator: str
    :type version2: str

    :rtype: bool
    """
    constraint = Constraint(operator, version2)

    return constraint.matches(Constraint('==', version1))



from pathlib import Path

from poetry.layouts import layout

from .command import Command


class NewCommand(Command):
    """
    Creates a new Python project at <path>

    new
        { path : The path to create the project at. }
        { --name : Set the resulting package name. }
    """

    def handle(self):
        layout_ = layout('standard')

        path = Path.cwd() / Path(self.argument('path'))
        name = self.option('name')
        if not name:
            name = path.name

        if path.exists():
            if list(path.glob('*')):
                # Directory is not empty. Aborting.
                raise RuntimeError(
                    'Destination <fg=yellow;bg=red>{}</> '
                    'exists and is not empty'.format(
                        path
                    )
                )

        readme_format = 'rst'

        layout_ = layout_(name, '0.1.0', readme_format=readme_format)
        layout_.create(path)

        self.line(
            'Created package <info>{}</> in <fg=blue>{}</>'
            .format(name, path.relative_to(Path.cwd()))
        )



from . import toplevels
from .cascadedict import CascadeDict


class NamedDict(dict):
    """
    A dict that can use Name instances as keys.
    """

    def __init__(self, other_dict=None):
        dict.__init__(self)
        if other_dict:
            for k, v in other_dict.items():
                self[k] = v

    def __setitem__(self, key, value):
        """
        key can be an Name instance.

        When key is a path in the form of an Name instance,
        all the parents and grandparents of the value are
        created along the way as instances of NamedDict.

        If the parent of the value exists, it is replaced with a
        CascadeDict() that cascades the old parent value
        with a new NamedDict that contains the given child name and value.
        """
        if isinstance(key, toplevels.Name):
            obj = self
            for i, name in enumerate(key.sub_names):
                if name in obj:
                    if i == len(key.sub_names) - 1:
                        obj[name] = CascadeDict(obj[name], value)
                    else:
                        obj[name] = CascadeDict(NamedDict(), obj[name])
                else:
                    if i == len(key.sub_names) - 1:
                        obj[name] = value
                    else:
                        obj[name] = NamedDict()

                obj = obj[name]
        else:
            return dict.__setitem__(self, key, value)

    def __contains__(self, item):
        try:
            _ = self[item]
            return True
        except KeyError:
            return False

    def append(self, key, value):
        """
        Makes sure the value pointed to by key exists
        and is a list and appends the given value to it.
        """
        if key in self:
            self[key].append(value)
        else:
            self[key] = [value]

    def __getitem__(self, item):
        if isinstance(item, toplevels.Name):
            d = self
            for name in item.sub_names:
                d = d[name]
            return d
        else:
            return dict.__getitem__(self, item)

    def __eq__(self, other):
        return dict.__eq__(self, other)


def structure(table_toplevels):
    """
    Accepts an ordered sequence of TopLevel instances and returns a navigable
    object structure representation of the TOML file.
    """
    table_toplevels = tuple(table_toplevels)
    obj = NamedDict()

    last_array_of_tables = None  # The Name of the last array-of-tables header

    for toplevel in table_toplevels:

        if isinstance(toplevel, toplevels.AnonymousTable):
            obj[''] = toplevel.table_element

        elif isinstance(toplevel, toplevels.Table):
            if last_array_of_tables and toplevel.name.is_prefixed_with(last_array_of_tables):
                seq = obj[last_array_of_tables]
                unprefixed_name = toplevel.name.without_prefix(last_array_of_tables)

                seq[-1] = CascadeDict(seq[-1], NamedDict({unprefixed_name: toplevel.table_element}))
            else:
                obj[toplevel.name] = toplevel.table_element
        else:    # It's an ArrayOfTables

            if last_array_of_tables and toplevel.name != last_array_of_tables and \
                    toplevel.name.is_prefixed_with(last_array_of_tables):

                seq = obj[last_array_of_tables]
                unprefixed_name = toplevel.name.without_prefix(last_array_of_tables)

                if unprefixed_name in seq[-1]:
                    seq[-1][unprefixed_name].append(toplevel.table_element)
                else:
                    cascaded_with = NamedDict({unprefixed_name: [toplevel.table_element]})
                    seq[-1] = CascadeDict(seq[-1], cascaded_with)

            else:
                obj.append(toplevel.name, toplevel.table_element)
                last_array_of_tables = toplevel.name

    return obj



import re

_modifier_regex = (
    '[._-]?'
    '(?:(stable|beta|b|RC|c|pre|alpha|a|patch|pl|p|post|[a-z])'
    '((?:[.-]?\d+)*)?)?'
    '([.-]?dev)?'
)


def normalize_version(version):
    """
    Normalizes a version string to be able to perform comparisons on it.
    """
    version = version.strip()

    # strip off build metadata
    m = re.match('^([^,\s+]+)\+[^\s]+$', version)
    if m:
        version = m.group(1)

    index = None
    # Match classic versioning
    m = re.match(
        '(?i)^v?(\d{{1,5}})(\.\d+)?(\.\d+)?(\.\d+)?{}$'.format(
            _modifier_regex
        ),
        version
    )
    if m:
        version = f'{m.group(1)}' \
                  f'{m.group(2) if m.group(2) else ".0"}' \
                  f'{m.group(3) if m.group(3) else ".0"}' \
                  f'{m.group(4) if m.group(4) else ".0"}'
        index = 5
    else:
        # Some versions have the form M.m.p-\d+
        # which means M.m.p-post\d+
        m = re.match(
            '(?i)^v?(\d{{1,5}})(\.\d+)?(\.\d+)?(\.\d+)?-(?:\d+){}$'.format(
                _modifier_regex
            ),
            version
        )
        if m:
            version = f'{m.group(1)}' \
                      f'{m.group(2) if m.group(2) else ".0"}' \
                      f'{m.group(3) if m.group(3) else ".0"}' \
                      f'{m.group(4) if m.group(4) else ".0"}'
            index = 5
        else:
            # Match date(time) based versioning
            m = re.match(
                '(?i)^v?(\d{{4}}(?:[.:-]?\d{{2}}){{1,6}}(?:[.:-]?\d{{1,3}})?){}$'.format(
                    _modifier_regex
                ),
                version
            )
            if m:
                version = re.sub('\D', '.', m.group(1))

                index = 2

    # add version modifiers if a version was matched
    if index is not None:
        if len(m.groups()) - 1 >= index and m.group(index):
            if m.group(index) == 'post':
                # Post releases should be considered
                # stable releases
                return version

            version = f'{version}' \
                      f'-{_expand_stability(m.group(index))}'

            if m.group(index + 1):
                version = f'{version}.{m.group(index + 1).lstrip(".-")}'

        return version

    raise ValueError(f'Invalid version string "{version}"')


def normalize_stability(stability: str) -> str:
    stability = stability.lower()

    if stability == 'rc':
        return 'RC'

    return stability


def parse_stability(version: str) -> str:
    """
    Returns the stability of a version.
    """
    version = re.sub('(?i)#.+$', '', version)

    if 'dev-' == version[:4] or '-dev' == version[-4:]:
        return 'dev'

    m = re.search('(?i){}(?:\+.*)?$'.format(_modifier_regex), version.lower())
    if m:
        if m.group(3):
            return 'dev'

        if m.group(1):
            if m.group(1) in ['beta', 'b']:
                return 'beta'
            elif m.group(1) in ['alpha', 'a']:
                return 'alpha'
            elif m.group(1) in ['rc', 'c']:
                return 'RC'
            elif m.group(1) == 'post':
                return 'stable'
            else:
                return 'dev'

    return 'stable'


def _expand_stability(stability: str) -> str:
    stability = stability.lower()

    if stability == 'a':
        return 'alpha'
    elif stability == 'b':
        return 'beta'
    elif stability in ['c', 'pre']:
        return 'rc'
    elif stability in ['p', 'pl']:
        return 'patch'
    elif stability in ['post']:
        return ''

    return stability



import contextlib
import hashlib
import os
import re
import tempfile
import shutil
import stat
import zipfile

from base64 import urlsafe_b64encode
from io import StringIO
from pathlib import Path
from types import SimpleNamespace

from poetry.__version__ import __version__
from poetry.semver.constraints import Constraint
from poetry.semver.constraints import MultiConstraint

from ..utils.helpers import normalize_file_permissions
from ..utils.tags import get_abbr_impl
from ..utils.tags import get_abi_tag
from ..utils.tags import get_impl_ver
from ..utils.tags import get_platform
from .builder import Builder


wheel_file_template = """\
Wheel-Version: 1.0
Generator: poetry {version}
Root-Is-Purelib: {pure_lib}
Tag: {tag}
"""


class WheelBuilder(Builder):

    def __init__(self, poetry, venv, io, target_fp, original=None):
        super().__init__(poetry, venv, io)

        self._records = []
        self._original_path = self._path
        if original:
            self._original_path = original.file.parent

        # Open the zip file ready to write
        self._wheel_zip = zipfile.ZipFile(target_fp, 'w',
                                          compression=zipfile.ZIP_DEFLATED)

    @classmethod
    def make_in(cls, poetry, venv, io, directory, original=None) -> SimpleNamespace:
        # We don't know the final filename until metadata is loaded, so write to
        # a temporary_file, and rename it afterwards.
        (fd, temp_path) = tempfile.mkstemp(suffix='.whl',
                                           dir=str(directory))
        try:
            with open(fd, 'w+b') as fp:
                wb = WheelBuilder(poetry, venv, io, fp, original=original)
                wb.build()

            wheel_path = directory / wb.wheel_filename
            os.replace(temp_path, str(wheel_path))
        except:
            os.unlink(temp_path)
            raise

        return SimpleNamespace(builder=wb, file=wheel_path)

    @classmethod
    def make(cls, poetry, venv, io) -> SimpleNamespace:
        """Build a wheel in the dist/ directory, and optionally upload it.
            """
        dist_dir = poetry.file.parent / 'dist'
        try:
            dist_dir.mkdir()
        except FileExistsError:
            pass

        return cls.make_in(poetry, venv, io, dist_dir)

    def build(self) -> None:
        self._io.writeln(' - Building <info>wheel</info>')
        try:
            self._build()
            self.copy_module()
            self.write_metadata()
            self.write_record()
        finally:
            self._wheel_zip.close()

        self._io.writeln(f' - Built <fg=cyan>{self.wheel_filename}</>')

    def _build(self) -> None:
        if self._package.build:
            setup = self._path / 'setup.py'

            # We need to place ourselves in the temporary
            # directory in order to build the package
            current_path = os.getcwd()
            try:
                os.chdir(str(self._path))
                self._venv.run(
                    'python',
                    str(setup),
                    'build',
                    '-b', str(self._path / 'build')
                )
            finally:
                os.chdir(current_path)

            build_dir = self._path / 'build'
            lib = list(build_dir.glob('lib.*'))
            if not lib:
                # The result of building the extensions
                # does not exist, this may due to conditional
                # builds, so we assume that it's okay
                return

            lib = lib[0]
            for pkg in lib.glob('*'):
                shutil.rmtree(str(self._path / pkg.name))
                shutil.copytree(str(pkg), str(self._path / pkg.name))

    def copy_module(self) -> None:
        if self._module.is_package():
            files = self.find_files_to_add()

            # Walk the files and compress them,
            # sorting everything so the order is stable.
            for file in sorted(files):
                full_path = self._path / file

                # Do not include topmost files
                if full_path.relative_to(self._path) == Path(file.name):
                    continue

                self._add_file(full_path, file)
        else:
            self._add_file(str(self._module.path), self._module.path.name)

    def write_metadata(self):
        if 'scripts' in self._poetry.config or 'plugins' in self._poetry.config:
            with self._write_to_zip(self.dist_info + '/entry_points.txt') as f:
                self._write_entry_points(f)

        for base in ('COPYING', 'LICENSE'):
            for path in sorted(self._path.glob(base + '*')):
                self._add_file(path, '%s/%s' % (self.dist_info, path.name))

        with self._write_to_zip(self.dist_info + '/WHEEL') as f:
            self._write_wheel_file(f)

        with self._write_to_zip(self.dist_info + '/METADATA') as f:
            self._write_metadata_file(f)

    def write_record(self):
        # Write a record of the files in the wheel
        with self._write_to_zip(self.dist_info + '/RECORD') as f:
            for path, hash, size in self._records:
                f.write('{},sha256={},{}\n'.format(path, hash, size))
            # RECORD itself is recorded with no hash or size
            f.write(self.dist_info + '/RECORD,,\n')

    def find_excluded_files(self) -> list:
        # Checking VCS
        return []

    @property
    def dist_info(self) -> str:
        return self.dist_info_name(self._package.name, self._package.version)

    @property
    def wheel_filename(self) -> str:
        return '{}-{}-{}.whl'.format(
            re.sub("[^\w\d.]+", "_", self._package.pretty_name, flags=re.UNICODE),
            re.sub("[^\w\d.]+", "_", self._package.version, flags=re.UNICODE),
            self.tag
        )

    def supports_python2(self):
        return self._package.python_constraint.matches(
            MultiConstraint([
                Constraint('>=', '2.0.0'),
                Constraint('<', '3.0.0')
            ])
        )

    def dist_info_name(self, distribution, version) -> str:
        escaped_name = re.sub("[^\w\d.]+", "_", distribution, flags=re.UNICODE)
        escaped_version = re.sub("[^\w\d.]+", "_", version, flags=re.UNICODE)

        return '{}-{}.dist-info'.format(escaped_name, escaped_version)

    @property
    def tag(self):
        if self._package.build:
            platform = get_platform().replace('.', '_').replace('-', '_')
            impl_name = get_abbr_impl(self._venv)
            impl_ver = get_impl_ver(self._venv)
            impl = impl_name + impl_ver
            abi_tag = str(get_abi_tag(self._venv)).lower()
            tag = (impl, abi_tag, platform)
        else:
            platform = 'any'
            if self.supports_python2():
                impl = 'py2.py3'
            else:
                impl = 'py3'

            tag = (impl, 'none', platform)

        return '-'.join(tag)

    def _add_file(self, full_path, rel_path):
        full_path, rel_path = str(full_path), str(rel_path)
        if os.sep != '/':
            # We always want to have /-separated paths in the zip file and in
            # RECORD
            rel_path = rel_path.replace(os.sep, '/')

        zinfo = zipfile.ZipInfo.from_file(full_path, rel_path)

        # Normalize permission bits to either 755 (executable) or 644
        st_mode = os.stat(full_path).st_mode
        new_mode = normalize_file_permissions(st_mode)
        zinfo.external_attr = (new_mode & 0xFFFF) << 16  # Unix attributes

        if stat.S_ISDIR(st_mode):
            zinfo.external_attr |= 0x10  # MS-DOS directory flag

        hashsum = hashlib.sha256()
        with open(full_path, 'rb') as src, self._wheel_zip.open(zinfo,
                                                                'w') as dst:
            while True:
                buf = src.read(1024 * 8)
                if not buf:
                    break
                hashsum.update(buf)
                dst.write(buf)

        size = os.stat(full_path).st_size
        hash_digest = urlsafe_b64encode(hashsum.digest()).decode(
            'ascii').rstrip('=')

        self._records.append((rel_path, hash_digest, size))

    @contextlib.contextmanager
    def _write_to_zip(self, rel_path):
        sio = StringIO()
        yield sio

        # The default is a fixed timestamp rather than the current time, so
        # that building a wheel twice on the same computer can automatically
        # give you the exact same result.
        date_time = (2016, 1, 1, 0, 0, 0)
        zi = zipfile.ZipInfo(rel_path, date_time)
        b = sio.getvalue().encode('utf-8')
        hashsum = hashlib.sha256(b)
        hash_digest = urlsafe_b64encode(
            hashsum.digest()
        ).decode('ascii').rstrip('=')

        self._wheel_zip.writestr(zi, b, compress_type=zipfile.ZIP_DEFLATED)
        self._records.append((rel_path, hash_digest, len(b)))

    def _write_entry_points(self, fp):
        """
        Write entry_points.txt.
        """
        entry_points = self.convert_entry_points()

        for group_name in sorted(entry_points):
            fp.write('[{}]\n'.format(group_name))
            for ep in sorted(entry_points[group_name]):
                fp.write(ep.replace(' ', '') + '\n')

            fp.write('\n')

    def _write_wheel_file(self, fp):
        fp.write(
            wheel_file_template.format(
                version=__version__,
                pure_lib='true' if self._package.build is None else 'false',
                tag=self.tag
            )
        )

    def _write_metadata_file(self, fp):
        """
        Write out metadata in the 2.x format (email like)
        """
        fp.write('Metadata-Version: 2.1\n')
        fp.write(f'Name: {self._meta.name}\n')
        fp.write(f'Version: {self._meta.version}\n')
        fp.write(f'Summary: {self._meta.summary}\n')
        fp.write(f'Home-page: {self._meta.home_page or "UNKNOWN"}\n')
        fp.write(f'License: {self._meta.license or "UNKOWN"}\n')

        # Optional fields
        if self._meta.keywords:
            fp.write(f"Keywords: {self._meta.keywords}\n")

        if self._meta.author:
            fp.write(f'Author: {self._meta.author}\n')

        if self._meta.author_email:
            fp.write(f'Author-email: {self._meta.author_email}\n')

        if self._meta.requires_python:
            fp.write(f'Requires-Python: {self._meta.requires_python}\n')

        for classifier in self._meta.classifiers:
            fp.write(f'Classifier: {classifier}\n')

        for dep in self._meta.requires_dist:
            fp.write(f'Requires-Dist: {dep}\n')

        if self._meta.description_content_type:
            fp.write(f'Description-Content-Type: '
                     f'{self._meta.description_content_type}\n')

        # TODO: Provides extra

        if self._meta.description is not None:
            fp.write('\n' + self._meta.description + '\n')



import os
import re

from poetry.semver.version_parser import VersionParser
from poetry.version.markers import Marker
from poetry.version.requirements import Requirement

from .dependency import Dependency
from .locker import Locker
from .package import Package
from .utils.link import Link
from .utils.utils import convert_markers
from .utils.utils import group_markers
from .utils.utils import is_archive_file
from .utils.utils import is_installable_dir
from .utils.utils import is_url
from .utils.utils import path_to_url
from .utils.utils import strip_extras
from .vcs_dependency import VCSDependency


def dependency_from_pep_508(name):
    req = Requirement(name)

    if req.marker:
        markers = convert_markers(req.marker.markers)
    else:
        markers = {}

    name = req.name
    path = os.path.normpath(os.path.abspath(name))
    link = None

    if is_url(name):
        link = Link(name)
    else:
        p, extras = strip_extras(path)
        if (os.path.isdir(p) and
                (os.path.sep in name or name.startswith('.'))):

            if not is_installable_dir(p):
                raise ValueError(
                    "Directory %r is not installable. File 'setup.py' "
                    "not found." % name
                )
            link = Link(path_to_url(p))
        elif is_archive_file(p):
            link = Link(path_to_url(p))

    # it's a local file, dir, or url
    if link:
        # Handle relative file URLs
        if link.scheme == 'file' and re.search(r'\.\./', link.url):
            link = Link(
                path_to_url(os.path.normpath(os.path.abspath(link.path)))
            )
        # wheel file
        if link.is_wheel:
            m = re.match(
                '^(?P<namever>(?P<name>.+?)-(?P<ver>\d.*?))',
                link.filename
            )
            if not m:
                raise ValueError(f'Invalid wheel name: {link.filename}')

            name = m.group('name')
            version = m.group('ver')
            dep = Dependency(name, version)
        else:
            name = link.egg_fragment

            if link.scheme == 'git':
                dep = VCSDependency(name, 'git', link.url_without_fragment)
            else:
                dep = Dependency(name, '*')
    else:
        if req.pretty_constraint:
            constraint = req.constraint
        else:
            constraint = '*'

        dep = Dependency(name, constraint)

    if 'extra' in markers:
        # If we have extras, the dependency is optional
        dep.deactivate()

        for or_ in markers['extra']:
            for _, extra in or_:
                dep.extras.append(extra)

    if 'python_version' in markers:
        ors = []
        for or_ in markers['python_version']:
            ands = []
            for op, version in or_:
                # Expand python version
                if op == '==':
                    version = '~' + version
                    op = ''
                elif op == '!=':
                    version += '.*'

                ands.append(f'{op}{version}')

            ors.append(' '.join(ands))

        dep.python_versions = ' || '.join(ors)

    if 'sys_platform' in markers:
        ors = []
        for or_ in markers['sys_platform']:
            ands = []
            for op, platform in or_:
                if op == '==':
                    op = ''

                ands.append(f'{op}{platform}')

            ors.append(' '.join(ands))

        dep.platform = ' || '.join(ors)

    return dep



from abc import abstractmethod

TYPE_METADATA = 'element-metadata'
TYPE_ATOMIC = 'element-atomic'
TYPE_CONTAINER = 'element-container'
TYPE_MARKUP = 'element-markup'


class Element:
    """
    An Element:
        - is one or more Token instances, or one or more other Element instances. Not both.
        - knows how to serialize its value back to valid TOML code.

    A non-metadata Element is an Element that:
        - knows how to deserialize its content into usable Python primitive, seq-like,  or dict-like value.
        - knows how to update its content from a Python primitive, seq-like, or dict-like value
            while maintaining its formatting.
    """

    def __init__(self, _type):
        self._type = _type

    @property
    def type(self):
        return self._type

    @abstractmethod
    def serialized(self):
        """
        TOML serialization of this element as str.
        """
        raise NotImplementedError


class TokenElement(Element):
    """
    An Element made up of tokens
    """

    def __init__(self, _tokens, _type):
        Element.__init__(self, _type)
        self._validate_tokens(_tokens)
        self._tokens = list(_tokens)

    @property
    def tokens(self):
        return self._tokens

    @property
    def first_token(self):
        return self._tokens[0]

    @abstractmethod
    def _validate_tokens(self, _tokens):
        raise NotImplementedError

    def serialized(self):
        return ''.join(token.source_substring for token in self._tokens)

    def __repr__(self):
        return repr(self.tokens)

    @property
    def primitive_value(self):
        """
        Returns a primitive Python value without any formatting or markup metadata.
        """
        raise NotImplementedError


class ContainerElement(Element):
    """
    An Element containing exclusively other elements.
    """

    def __init__(self, sub_elements):
        Element.__init__(self, TYPE_CONTAINER)
        self._sub_elements = list(sub_elements)

    @property
    def sub_elements(self):
        return self._sub_elements

    @property
    def elements(self):
        return self.sub_elements

    def serialized(self):
        return ''.join(element.serialized() for element in self.sub_elements)

    def __eq__(self, other):
        return self.primitive_value == other

    def __repr__(self):
        return repr(self.primitive_value)

    @property
    def primitive_value(self):
        """
        Returns a primitive Python value without any formatting or markup metadata.
        """
        raise NotImplementedError





from cleo.styles import CleoStyle
from cleo.helpers import ProgressIndicator
from poetry.mixology.contracts import UI as BaseUI


class UI(BaseUI):

    def __init__(self, io: CleoStyle):
        self._io = io
        self._progress = None

        super().__init__(self._io.is_debug())

    @property
    def output(self):
        return self._io

    def before_resolution(self) -> None:
        self._io.write('<info>Resolving dependencies</>')

        if self.is_debugging():
            self._io.new_line()

    def indicate_progress(self):
        if not self.is_debugging():
            self._io.write('.')

    def after_resolution(self) -> None:
        self._io.new_line()

    def debug(self, message, depth) -> None:
        if self.is_debugging():
            debug_info = str(message)
            debug_info = '\n'.join([
                '<comment>:{}:</> {}'.format(str(depth).rjust(4), s)
                for s in debug_info.split('\n')
            ]) + '\n'

            self.output.write(debug_info)



from .action import Action
from .vertex import Vertex

_NULL = object()


class AddVertex(Action):

    def __init__(self, name, payload, root):
        """
        :param name: The name of the vertex.
        :type name: str

        :param payload: The payload of he vertex
        :type payload: Any

        :param root: whether the vertex is root or not
        :type root: bool
        """
        super(AddVertex, self).__init__()

        self._name = name
        self._payload = payload
        self._root = root
        self._existing_payload = _NULL
        self._existing_root = None

    @property
    def action_name(self):
        return 'add_vertex'

    @property
    def name(self):
        return self._name

    @property
    def payload(self):
        return self._payload

    @property
    def root(self):
        return self._root

    def up(self, graph):
        existing = graph.vertices.get(self._name)
        if existing:
            self._existing_payload = existing.payload
            self._existing_root = existing.root

        vertex = existing or Vertex(self._name, self._payload)
        graph.vertices[vertex.name] = vertex
        if not vertex.payload:
            vertex.payload = self.payload

        if not vertex.root:
            vertex.root = self.root

        return vertex

    def down(self, graph):
        if self._existing_payload is not _NULL:
            vertex = graph.vertices[self._name]
            vertex.payload = self._existing_payload
            vertex.root = self._existing_root
        else:
            del graph.vertices[self._name]



import re

_canonicalize_regex = re.compile('[-_.]+')


def canonicalize_name(name: str) -> str:
    return _canonicalize_regex.sub('-', name).lower()


def module_name(name: str) -> str:
    return canonicalize_name(name).replace('-', '_')



from .base_constraint import BaseConstraint


class EmptyConstraint(BaseConstraint):

    pretty_string = None

    def matches(self, _):
        return True

    def __str__(self):
        return '*'



import re

from collections import namedtuple
from itertools import dropwhile

from .base import BaseVersion
from .exceptions import InvalidVersion
from .utils import Infinity


_Version = namedtuple(
    "_Version",
    ["epoch", "release", "dev", "pre", "post", "local"],
)


VERSION_PATTERN = re.compile("""
    ^
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_.]?
                (?P<post_l>post|rev|r)
                [-_.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_.]?
            (?P<dev_l>dev)
            [-_.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_.][a-z0-9]+)*))?       # local version
    $
""", re.IGNORECASE | re.VERBOSE)


class Version(BaseVersion):

    def __init__(self, version):
        # Validate the version and parse it into pieces
        match = VERSION_PATTERN.match(version)
        if not match:
            raise InvalidVersion("Invalid version: '{0}'".format(version))

        # Store the parsed out pieces of the version
        self._version = _Version(
            epoch=int(match.group("epoch")) if match.group("epoch") else 0,
            release=tuple(int(i) for i in match.group("release").split(".")),
            pre=_parse_letter_version(
                match.group("pre_l"),
                match.group("pre_n"),
            ),
            post=_parse_letter_version(
                match.group("post_l"),
                match.group("post_n1") or match.group("post_n2"),
            ),
            dev=_parse_letter_version(
                match.group("dev_l"),
                match.group("dev_n"),
            ),
            local=_parse_local_version(match.group("local")),
        )

        # Generate a key which will be used for sorting
        self._key = _cmpkey(
            self._version.epoch,
            self._version.release,
            self._version.pre,
            self._version.post,
            self._version.dev,
            self._version.local,
        )

    def __repr__(self):
        return "<Version({0})>".format(repr(str(self)))

    def __str__(self):
        parts = []

        # Epoch
        if self._version.epoch != 0:
            parts.append("{0}!".format(self._version.epoch))

        # Release segment
        parts.append(".".join(str(x) for x in self._version.release))

        # Pre-release
        if self._version.pre is not None:
            parts.append("".join(str(x) for x in self._version.pre))

        # Post-release
        if self._version.post is not None:
            parts.append(".post{0}".format(self._version.post[1]))

        # Development release
        if self._version.dev is not None:
            parts.append(".dev{0}".format(self._version.dev[1]))

        # Local version segment
        if self._version.local is not None:
            parts.append(
                "+{0}".format(".".join(str(x) for x in self._version.local))
            )

        return "".join(parts)

    @property
    def public(self):
        return str(self).split("+", 1)[0]

    @property
    def base_version(self):
        parts = []

        # Epoch
        if self._version.epoch != 0:
            parts.append("{0}!".format(self._version.epoch))

        # Release segment
        parts.append(".".join(str(x) for x in self._version.release))

        return "".join(parts)

    @property
    def local(self):
        version_string = str(self)
        if "+" in version_string:
            return version_string.split("+", 1)[1]

    @property
    def is_prerelease(self):
        return bool(self._version.dev or self._version.pre)

    @property
    def is_postrelease(self):
        return bool(self._version.post)


def _parse_letter_version(letter, number):
    if letter:
        # We consider there to be an implicit 0 in a pre-release if there is
        # not a numeral associated with it.
        if number is None:
            number = 0

        # We normalize any letters to their lower case form
        letter = letter.lower()

        # We consider some words to be alternate spellings of other words and
        # in those cases we want to normalize the spellings to our preferred
        # spelling.
        if letter == "alpha":
            letter = "a"
        elif letter == "beta":
            letter = "b"
        elif letter in ["c", "pre", "preview"]:
            letter = "rc"
        elif letter in ["rev", "r"]:
            letter = "post"

        return letter, int(number)
    if not letter and number:
        # We assume if we are given a number, but we are not given a letter
        # then this is using the implicit post release syntax (e.g. 1.0-1)
        letter = "post"

        return letter, int(number)


_local_version_seperators = re.compile(r"[._-]")


def _parse_local_version(local):
    """
    Takes a string like abc.1.twelve and turns it into ("abc", 1, "twelve").
    """
    if local is not None:
        return tuple(
            part.lower() if not part.isdigit() else int(part)
            for part in _local_version_seperators.split(local)
        )


def _cmpkey(epoch, release, pre, post, dev, local):
    # When we compare a release version, we want to compare it with all of the
    # trailing zeros removed. So we'll use a reverse the list, drop all the now
    # leading zeros until we come to something non zero, then take the rest
    # re-reverse it back into the correct order and make it a tuple and use
    # that for our sorting key.
    release = tuple(
        reversed(list(
            dropwhile(
                lambda x: x == 0,
                reversed(release),
            )
        ))
    )

    # We need to "trick" the sorting algorithm to put 1.0.dev0 before 1.0a0.
    # We'll do this by abusing the pre segment, but we _only_ want to do this
    # if there is not a pre or a post segment. If we have one of those then
    # the normal sorting rules will handle this case correctly.
    if pre is None and post is None and dev is not None:
        pre = -Infinity

    # Versions without a pre-release (except as noted above) should sort after
    # those with one.
    elif pre is None:
        pre = Infinity

    # Versions without a post segment should sort before those with one.
    if post is None:
        post = -Infinity

    # Versions without a development segment should sort after those with one.
    if dev is None:
        dev = Infinity

    if local is None:
        # Versions without a local segment should sort before those with one.
        local = -Infinity
    else:
        # Versions with a local segment need that segment parsed to implement
        # the sorting rules in PEP440.
        # - Alpha numeric segments sort before numeric segments
        # - Alpha numeric segments sort lexicographically
        # - Numeric segments sort numerically
        # - Shorter versions sort before longer versions when the prefixes
        #   match exactly
        local = tuple(
            (i, "") if isinstance(i, int) else (-Infinity, i)
            for i in local
        )

    return epoch, release, pre, post, dev, local



from pathlib import Path
from typing import List
from typing import Union

from cachy import CacheManager
from requests import get

from poetry.locations import CACHE_DIR
from poetry.packages import dependency_from_pep_508
from poetry.packages import Package
from poetry.semver.constraints import Constraint
from poetry.semver.constraints.base_constraint import BaseConstraint
from poetry.semver.version_parser import VersionParser
from poetry.version.markers import InvalidMarker

from .repository import Repository


class PyPiRepository(Repository):

    def __init__(self, url='https://pypi.org/', disable_cache=False):
        self._url = url
        self._disable_cache = disable_cache
        self._cache = CacheManager({
            'default': 'releases',
            'serializer': 'json',
            'stores': {
                'releases': {
                    'driver': 'file',
                    'path': Path(CACHE_DIR) / 'cache' / 'repositories' / 'pypi'
                },
                'packages': {
                    'driver': 'dict'
                }
            }
        })
        
        super().__init__()

    def find_packages(self,
                      name: str,
                      constraint: Union[Constraint, str, None] = None,
                      extras: Union[list, None] = None
                      ) -> List[Package]:
        """
        Find packages on the remote server.
        """
        packages = []

        if constraint is not None and not isinstance(constraint, BaseConstraint):
            version_parser = VersionParser()
            constraint = version_parser.parse_constraints(constraint)

        info = self.get_package_info(name)

        versions = []

        for version, release in info['releases'].items():
            if (
                not constraint
                or (constraint and constraint.matches(Constraint('=', version)))
            ):
                versions.append(version)

        for version in versions:
            packages.append(
                self.package(name, version, extras=extras)
            )

        return packages

    def package(self,
                name: str,
                version: str,
                extras: Union[list, None] = None) -> Package:
        try:
            index = self._packages.index(Package(name, version, version))

            return self._packages[index]
        except ValueError:
            if extras is None:
                extras = []

            release_info = self.get_release_info(name, version)
            package = Package(name, version, version)
            requires_dist = release_info['requires_dist'] or []
            for req in requires_dist:
                try:
                    dependency = dependency_from_pep_508(req)
                except InvalidMarker:
                    # Invalid marker
                    # We strip the markers hoping for the best
                    req = req.split(';')[0]

                    dependency = dependency_from_pep_508(req)

                if dependency.extras:
                    for extra in dependency.extras:
                        if extra not in package.extras:
                            package.extras[extra] = []

                        package.extras[extra].append(dependency)

                if not dependency.is_optional():
                    package.requires.append(dependency)

            # Adding description
            package.description = release_info.get('summary', '')

            # Adding hashes information
            package.hashes = release_info['digests']

            # Activate extra dependencies
            for extra in extras:
                if extra in package.extras:
                    for dep in package.extras[extra]:
                        dep.activate()

                    package.requires += package.extras[extra]

            self._packages.append(package)

            return package

    def search(self, query, mode=0):
        results = []

        search = {
            'name': query
        }

        if mode == self.SEARCH_FULLTEXT:
            search['summary'] = query

        client = ServerProxy(self._url)
        hits = client.search(search, 'or')

        for hit in hits:
            results.append({
                'name': hit['name'],
                'description': hit['summary'],
                'version': hit['version']
            })

        return results

    def get_package_info(self, name: str) -> dict:
        """
        Return the package information given its name.

        The information is returned from the cache if it exists
        or retrieved from the remote server.
        """
        if self._disable_cache:
            return self._get_package_info(name)

        return self._cache.store('packages').remember_forever(
            f'{name}',
            lambda: self._get_package_info(name)
        )

    def _get_package_info(self, name: str) -> dict:
        data = self._get(self._url + f'pypi/{name}/json')
        if data is None:
            raise ValueError(f'Package [{name}] not found.')

        return data

    def get_release_info(self, name: str, version: str) -> dict:
        """
        Return the release information given a package name and a version.

        The information is returned from the cache if it exists
        or retrieved from the remote server.
        """
        if self._disable_cache:
            return self._get_release_info(name, version)

        return self._cache.remember_forever(
            f'{name}:{version}',
            lambda: self._get_release_info(name, version)
        )

    def _get_release_info(self, name: str, version: str) -> dict:
        json_data = self._get(self._url + f'pypi/{name}/{version}/json')
        if json_data is None:
            raise ValueError(f'Package [{name}] not found.')

        info = json_data['info']
        data = {
            'name': info['name'],
            'version': info['version'],
            'summary': info['summary'],
            'platform': info['platform'],
            'requires_dist': info['requires_dist'],
            'requires_python': info['requires_python'],
            'digests': []
        }
        for file_info in json_data['releases'][version]:
            data['digests'].append(file_info['digests']['sha256'])

        return data

    def _get(self, url: str) -> Union[dict, None]:
        json_response = get(url)
        if json_response.status_code == 404:
            return None

        json_data = json_response.json()

        return json_data



from .venv_command import VenvCommand

from poetry.masonry import Builder


class BuildCommand(VenvCommand):
    """
    Builds a package, as a tarball and a wheel by default.

    build
        { --f|format= : Limit the format to either wheel or sdist. }
    """

    def handle(self):
        fmt = 'all'
        if self.option('format'):
            fmt = self.option('format')

        package = self.poetry.package
        self.line(f'Building <info>{package.pretty_name}</> '
                  f'(<comment>{package.version}</>)')

        builder = Builder(self.poetry, self.venv, self.output)
        builder.build(fmt)



from poetry.installation import Installer

from .venv_command import VenvCommand


class RemoveCommand(VenvCommand):
    """
    Removes a package from the project dependencies.

    remove
        { packages* : Packages that should be removed. }
        {--D|dev : Removes a package from the development dependencies. }
        {--dry-run : Outputs the operations but will not execute anything
                     (implicitly enables --verbose). }
    """

    help = """The <info>remove</info> command removes a package from the current
list of installed packages

<info>poetry remove</info>"""

    def handle(self):
        packages = self.argument('packages')
        is_dev = self.option('dev')

        original_content = self.poetry.file.read()
        content = self.poetry.file.read()
        poetry_content = content['tool']['poetry']
        section = 'dependencies'
        if is_dev:
            section = 'dev-dependencies'

        # Deleting entries
        requirements = {}
        for name in packages:
            found = False
            for key in poetry_content[section]:
                if key.lower() == name.lower():
                    found = True
                    requirements[name] = poetry_content[section][name]
                    break

            if not found:
                raise ValueError(f'Package {name} not found')

        for key in requirements:
            del poetry_content[section][key]

        # Write the new content back
        self.poetry.file.write(content)

        # Update packages
        self.reset_poetry()

        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool
        )

        installer.dry_run(self.option('dry-run'))
        installer.update(True)
        installer.whitelist(requirements)

        try:
            status = installer.run()
        except Exception:
            self.poetry.file.write(original_content)

            raise

        if status != 0 or self.option('dry-run'):
            # Revert changes
            if not self.option('dry-run'):
                self.error(
                    '\n'
                    'Removal failed, reverting pyproject.toml '
                    'to its original content.'
                )

            self.poetry.file.write(original_content)

        return status



import re

from typing import List
from typing import Tuple

from poetry.installation import Installer
from poetry.semver.version_parser import VersionParser
from poetry.version.version_selector import VersionSelector

from .venv_command import VenvCommand


class AddCommand(VenvCommand):
    """
    Add a new dependency to <comment>pyproject.toml</>.

    add
        { name* : Packages to add. }
        {--D|dev : Add package as development dependency. }
        {--optional : Add as an optional dependency. }
        {--dry-run : Outputs the operations but will not execute anything
                     (implicitly enables --verbose). }
    """

    help = """The add command adds required packages to your <comment>pyproject.toml</> and installs them.

If you do not specify a version constraint, poetry will choose a suitable one based on the available package versions.
"""

    def handle(self):
        packages = self.argument('name')
        is_dev = self.option('dev')

        section = 'dependencies'
        if is_dev:
            section = 'dev-dependencies'

        original_content = self.poetry.file.read()
        content = self.poetry.file.read()
        poetry_content = content['tool']['poetry']

        for name in packages:
            for key in poetry_content[section]:
                if key.lower() == name.lower():
                    raise ValueError(f'Package {name} is already present')

        requirements = self._determine_requirements(packages)
        requirements = self._format_requirements(requirements)

        # validate requirements format
        parser = VersionParser()
        for constraint in requirements.values():
            parser.parse_constraints(constraint)

        for name, constraint in requirements.items():
            poetry_content[section][name] = constraint

        # Write new content
        self.poetry.file.write(content)

        # Cosmetic new line
        self.line('')

        # Update packages
        self.reset_poetry()

        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool
        )

        installer.dry_run(self.option('dry-run'))
        installer.update(True)
        installer.whitelist(requirements)

        try:
            status = installer.run()
        except Exception:
            self.poetry.file.write(original_content)

            raise

        if status != 0 or self.option('dry-run'):
            # Revert changes
            if not self.option('dry-run'):
                self.error(
                    '\n'
                    'Addition failed, reverting pyproject.toml '
                    'to its original content.'
                )

            self.poetry.file.write(original_content)

        return status

    def _determine_requirements(self, requires: List[str]) -> List[str]:
        if not requires:
            return []

        requires = self._parse_name_version_pairs(requires)
        result = []
        for requirement in requires:
            if 'version' not in requirement:
                # determine the best version automatically
                name, version = self._find_best_version_for_package(
                    requirement['name']
                )
                requirement['version'] = version
                requirement['name'] = name

                self.line(
                    f'Using version <info>{version}</> for <info>{name}</>'
                )
            else:
                # check that the specified version/constraint exists
                # before we proceed
                name, _ = self._find_best_version_for_package(
                    requirement['name'], requirement['version']
                )

                requirement['name'] = name

            result.append(f'{requirement["name"]} {requirement["version"]}')

        return result

    def _find_best_version_for_package(self,
                                       name,
                                       required_version=None
                                       ) -> Tuple[str, str]:
        selector = VersionSelector(self.poetry.pool)
        package = selector.find_best_candidate(name, required_version)

        if not package:
            # TODO: find similar
            raise ValueError(
                f'Could not find a matching version of package {name}'
            )

        return (
            package.pretty_name,
            selector.find_recommended_require_version(package)
        )

    def _parse_name_version_pairs(self, pairs: list) -> list:
        result = []

        for i in range(len(pairs)):
            pair = re.sub('^([^=: ]+)[=: ](.*)$', '\\1 \\2', pairs[i].strip())
            pair = pair.strip()

            if ' ' in pair:
                name, version = pair.split(' ', 2)
                result.append({
                    'name': name,
                    'version': version
                })
            else:
                result.append({
                    'name': pair
                })

        return result

    def _format_requirements(self, requirements: List[str]) -> dict:
        requires = {}
        requirements = self._parse_name_version_pairs(requirements)
        for requirement in requirements:
            requires[requirement['name']] = requirement['version']

        return requires



import re
from typing import Union

from poetry.semver.constraints import Constraint
from poetry.semver.constraints import EmptyConstraint
from poetry.semver.helpers import parse_stability
from poetry.semver.version_parser import VersionParser
from poetry.version import parse as parse_version

from.constraints.platform_constraint import PlatformConstraint
from .dependency import Dependency
from .vcs_dependency import VCSDependency

AUTHOR_REGEX = re.compile('(?u)^(?P<name>[- .,\w\d\'’"()]+) <(?P<email>.+?)>$')


class Package:

    AVAILABLE_PYTHONS = {
        '2',
        '2.7',
        '3',
        '3.4', '3.5', '3.6', '3.7'
    }

    supported_link_types = {
        'require': {
            'description': 'requires',
            'method': 'requires'
        },
        'provide': {
            'description': 'provides',
            'method': 'provides'
        }
    }

    STABILITY_STABLE = 0
    STABILITY_RC = 5
    STABILITY_BETA = 10
    STABILITY_ALPHA = 15
    STABILITY_DEV = 20

    stabilities = {
        'stable': STABILITY_STABLE,
        'rc': STABILITY_RC,
        'beta': STABILITY_BETA,
        'alpha': STABILITY_ALPHA,
        'dev': STABILITY_DEV,
    }

    def __init__(self, name, version, pretty_version=None):
        """
        Creates a new in memory package.
        """
        self._pretty_name = name
        self._name = name.lower()

        self._version = str(parse_version(version))
        self._pretty_version = pretty_version or version

        self.description = ''

        self._stability = parse_stability(version)
        self._dev = self._stability == 'dev'

        self._authors = []

        self.homepage = None
        self.repository_url = None
        self.keywords = []
        self.license = None
        self.readme = None

        self.source_type = ''
        self.source_reference = ''
        self.source_url = ''

        self.requires = []
        self.dev_requires = []
        self.extras = {}

        self._parser = VersionParser()

        self.category = 'main'
        self.hashes = []
        self.optional = False

        # Requirements for making it mandatory
        self.requirements = {}

        self.build = None
        self.include = []
        self.exclude = []

        self._python_versions = '*'
        self._python_constraint = self._parser.parse_constraints('*')
        self._platform = '*'
        self._platform_constraint = EmptyConstraint()

    @property
    def name(self):
        return self._name

    @property
    def pretty_name(self):
        return self._pretty_name

    @property
    def version(self):
        return self._version

    @property
    def pretty_version(self):
        return self._pretty_version

    @property
    def unique_name(self):
        return self.name + '-' + self._version

    @property
    def pretty_string(self):
        return self.pretty_name + ' ' + self.pretty_version
    
    @property
    def full_pretty_version(self):
        if not self._dev and self.source_type not in ['hg', 'git']:
            return self._pretty_version

        # if source reference is a sha1 hash -- truncate
        if len(self.source_reference) == 40:
            return '{} {}'.format(self._pretty_version,
                                  self.source_reference[0:7])

        return '{} {}'.format(self._pretty_version, self.source_reference)

    @property
    def authors(self) -> list:
        return self._authors

    @property
    def author_name(self) -> str:
        return self._get_author()['name']

    @property
    def author_email(self) -> str:
        return self._get_author()['email']

    def _get_author(self) -> dict:
        if not self._authors:
            return {
                'name': None,
                'email': None
            }

        m = AUTHOR_REGEX.match(self._authors[0])

        name = m.group('name')
        email = m.group('email')

        return {
            'name': name,
            'email': email
        }

    @property
    def python_versions(self):
        return self._python_versions

    @python_versions.setter
    def python_versions(self, value: str):
        self._python_versions = value
        self._python_constraint = self._parser.parse_constraints(value)

    @property
    def python_constraint(self):
        return self._python_constraint

    @property
    def platform(self) -> str:
        return self._platform

    @platform.setter
    def platform(self, value: str):
        self._platform = value
        self._platform_constraint = PlatformConstraint.parse(value)

    @property
    def platform_constraint(self):
        return self._platform_constraint
    
    @property
    def classifiers(self):
        classifiers = []

        # Automatically set python classifiers
        parser = VersionParser()
        if self.python_versions == '*':
            python_constraint = parser.parse_constraints('~2.7 || ^3.4')
        else:
            python_constraint = self.python_constraint

        for version in sorted(self.AVAILABLE_PYTHONS):
            if len(version) == 1:
                constraint = parser.parse_constraints(version + '.*')
            else:
                constraint = Constraint('=', version)

            if python_constraint.matches(constraint):
                classifiers.append(
                    f'Programming Language :: Python :: {version}'
                )

        return classifiers

    def is_dev(self):
        return self._dev

    def is_prerelease(self):
        return self._stability != 'stable'

    def add_dependency(self,
                       name: str,
                       constraint: Union[str, dict, None] = None,
                       category: str = 'main') -> Dependency:
        if constraint is None:
            constraint = '*'

        if isinstance(constraint, dict):
            if 'git' in constraint:
                # VCS dependency
                optional = constraint.get('optional', False)
                python_versions = constraint.get('python')
                platform = constraint.get('platform')

                dependency = VCSDependency(
                    name,
                    'git', constraint['git'],
                    branch=constraint.get('branch', None),
                    tag=constraint.get('tag', None),
                    rev=constraint.get('rev', None),
                    optional=optional,
                )

                if python_versions:
                    dependency.python_versions = python_versions

                if platform:
                    dependency.platform = platform
            else:
                version = constraint['version']
                optional = constraint.get('optional', False)
                allows_prereleases = constraint.get('allows_prereleases', False)
                python_versions = constraint.get('python')
                platform = constraint.get('platform')

                dependency = Dependency(
                    name, version,
                    optional=optional,
                    category=category,
                    allows_prereleases=allows_prereleases
                )

                if python_versions:
                    dependency.python_versions = python_versions

                if platform:
                    dependency.platform = platform

            if 'extras' in constraint:
                for extra in constraint['extras']:
                    dependency.extras.append(extra)
        else:
            dependency = Dependency(name, constraint, category=category)

        if category == 'dev':
            self.dev_requires.append(dependency)
        else:
            self.requires.append(dependency)

        return dependency

    def __hash__(self):
        return hash((self._name, self._version))

    def __eq__(self, other):
        if not isinstance(other, Package):
            return NotImplemented

        return self._name == other.name and self._version == other.version

    def __str__(self):
        return self.unique_name

    def __repr__(self):
        return '<Package {}>'.format(self.unique_name)




"""
A converter of python values to TOML Token instances.
"""
import codecs
import datetime
import six
import re

from .. import tokens
from ..errors import TOMLError
from ..tokens import Token
from ..util import chunkate_string


class NotPrimitiveError(TOMLError):
    pass


_operator_tokens_by_type = {
    tokens.TYPE_OP_SQUARE_LEFT_BRACKET: tokens.Token(tokens.TYPE_OP_SQUARE_LEFT_BRACKET, u'['),
    tokens.TYPE_OP_SQUARE_RIGHT_BRACKET: tokens.Token(tokens.TYPE_OP_SQUARE_RIGHT_BRACKET, u']'),
    tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET: tokens.Token(tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET, u'[['),
    tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET: tokens.Token(tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET, u']]'),
    tokens.TYPE_OP_COMMA: tokens.Token(tokens.TYPE_OP_COMMA, u','),
    tokens.TYPE_NEWLINE: tokens.Token(tokens.TYPE_NEWLINE, u'\n'),
    tokens.TYPE_OPT_DOT: tokens.Token(tokens.TYPE_OPT_DOT, u'.'),
}


def operator_token(token_type):
    return _operator_tokens_by_type[token_type]


def create_primitive_token(value, multiline_strings_allowed=True):
    """
    Creates and returns a single token for the given primitive atomic value.

    Raises NotPrimitiveError when the given value is not a primitive atomic value
    """
    if value is None:
        return create_primitive_token('')
    elif isinstance(value, bool):
        return tokens.Token(tokens.TYPE_BOOLEAN, u'true' if value else u'false')
    elif isinstance(value, int):
        return tokens.Token(tokens.TYPE_INTEGER, u'{}'.format(value))
    elif isinstance(value, float):
        return tokens.Token(tokens.TYPE_FLOAT, u'{}'.format(value))
    elif isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
        return tokens.Token(tokens.TYPE_DATE, strict_rfc3339.timestamp_to_rfc3339_utcoffset(ts))
    elif isinstance(value, six.string_types):
        return create_string_token(value, multiline_strings_allowed=multiline_strings_allowed)

    raise NotPrimitiveError("{} of type {}".format(value, type(value)))


_bare_string_regex = re.compile('^[a-zA-Z0-9_-]*$')


def create_string_token(text, bare_string_allowed=False, multiline_strings_allowed=True):
    """
    Creates and returns a single string token.

    Raises ValueError on non-string input.
    """
    if not isinstance(text, six.string_types):
        raise ValueError('Given value must be a string')

    if text == '':
        return tokens.Token(tokens.TYPE_STRING, '""'.format(_escape_single_line_quoted_string(text)))
    elif bare_string_allowed and _bare_string_regex.match(text):
        return tokens.Token(tokens.TYPE_BARE_STRING, text)
    elif multiline_strings_allowed and (len(tuple(c for c in text if c == '\n')) >= 2 or len(text) > 80):
        # If containing two or more newlines or is longer than 80 characters we'll use the multiline string format
        return _create_multiline_string_token(text)
    else:
        return tokens.Token(tokens.TYPE_STRING, '"{}"'.format(_escape_single_line_quoted_string(text)))


def _escape_single_line_quoted_string(text):
    if six.PY2:
        return text.encode('unicode-escape').encode('string-escape').replace('"', '\\"').replace("\\'", "'")
    else:
        return codecs.encode(text, 'unicode-escape').decode().replace('"', '\\"')


def _create_multiline_string_token(text):
    escaped = text.replace(u'"""', u'\"\"\"')
    if len(escaped) > 50:
        return tokens.Token(tokens.TYPE_MULTILINE_STRING, u'"""\n{}\\\n"""'.format(_break_long_text(escaped)))
    else:
        return tokens.Token(tokens.TYPE_MULTILINE_STRING, u'"""{}"""'.format(escaped))


def _break_long_text(text, maximum_length=75):
    """
    Breaks into lines of 75 character maximum length that are terminated by a backslash.
    """

    def next_line(remaining_text):

        # Returns a line and the remaining text

        if '\n' in remaining_text and remaining_text.index('\n') < maximum_length:
            i = remaining_text.index('\n')
            return remaining_text[:i+1], remaining_text[i+2:]
        elif len(remaining_text) > maximum_length and ' ' in remaining_text:
            i = remaining_text[:maximum_length].rfind(' ')
            return remaining_text[:i+1] + '\\\n', remaining_text[i+2:]
        else:
            return remaining_text, ''

    remaining_text = text
    lines = []
    while remaining_text:
        line, remaining_text = next_line(remaining_text)
        lines += [line]

    return ''.join(lines)


def create_whitespace(source_substring):
    return Token(tokens.TYPE_WHITESPACE, source_substring)


def create_multiline_string(text, maximum_line_length=120):
    def escape(t):
        return t.replace(u'"""', six.u(r'\"\"\"'))
    source_substring = u'"""\n{}"""'.format(u'\\\n'.join(chunkate_string(escape(text), maximum_line_length)))
    return Token(tokens.TYPE_MULTILINE_STRING, source_substring)



import re

from typing import List

from poetry.packages import Dependency
from poetry.puzzle import Solver
from poetry.repositories.repository import Repository
from poetry.semver.version_parser import VersionParser

from ..command import Command


class DebugResolveCommand(Command):
    """
    Debugs dependency resolution.

    debug:resolve
        { package?* : packages to resolve. }
    """

    def handle(self):
        packages = self.argument('package')

        if not packages:
            package = self.poetry.package
            dependencies = package.requires + package.dev_requires
        else:
            requirements = self._determine_requirements(packages)
            requirements = self._format_requirements(requirements)

            # validate requirements format
            parser = VersionParser()
            for constraint in requirements.values():
                parser.parse_constraints(constraint)

            dependencies = []
            for name, constraint in requirements.items():
                dependencies.append(
                    Dependency(name, constraint)
                )

        solver = Solver(
            self.poetry.package,
            self.poetry.pool,
            Repository(),
            self.output
        )

        ops = solver.solve(dependencies)

        self.line('')
        self.line('Resolution results:')
        self.line('')

        for op in ops:
            package = op.package
            self.line(f'  - <info>{package.name}</info> '
                      f'(<comment>{package.version}</comment>)')

    def _determine_requirements(self, requires: List[str]) -> List[str]:
        if not requires:
            return []

        requires = self._parse_name_version_pairs(requires)
        result = []
        for requirement in requires:
            if 'version' not in requirement:
                requirement['version'] = '*'

            result.append(f'{requirement["name"]} {requirement["version"]}')

        return result

    def _parse_name_version_pairs(self, pairs: list) -> list:
        result = []

        for i in range(len(pairs)):
            pair = re.sub('^([^=: ]+)[=: ](.*)$', '\\1 \\2', pairs[i].strip())
            pair = pair.strip()

            if ' ' in pair:
                name, version = pair.split(' ', 2)
                result.append({
                    'name': name,
                    'version': version
                })
            else:
                result.append({
                    'name': pair
                })

        return result

    def _format_requirements(self, requirements: List[str]) -> dict:
        requires = {}
        requirements = self._parse_name_version_pairs(requirements)
        for requirement in requirements:
            requires[requirement['name']] = requirement['version']

        return requires



import operator

from typing import Union

from .exceptions import InvalidVersion
from .legacy_version import LegacyVersion
from .version import Version


OP_EQ = operator.eq
OP_LT = operator.lt
OP_LE = operator.le
OP_GT = operator.gt
OP_GE = operator.ge
OP_NE = operator.ne

_trans_op = {
    '=': OP_EQ,
    '==': OP_EQ,
    '<': OP_LT,
    '<=': OP_LE,
    '>': OP_GT,
    '>=': OP_GE,
    '!=': OP_NE
}


def parse(version: str, strict: bool = False) -> Union[Version, LegacyVersion]:
    """
    Parse the given version string and return either a :class:`Version` object
    or a LegacyVersion object depending on if the given version is
    a valid PEP 440 version or a legacy version.

    If strict=True only PEP 440 versions will be accepted.
    """
    try:
        return Version(version)
    except InvalidVersion:
        if strict:
            raise

        return LegacyVersion(version)


def version_compare(version1: str, version2: str, operator) -> bool:
    if operator in _trans_op:
        operator = _trans_op[operator]
    elif operator in _trans_op.values():
        pass
    else:
        raise ValueError('Invalid operator')

    version1 = parse(version1)
    version2 = parse(version2)

    return operator(version1, version2)



from poetry.packages import Package
from poetry.utils.venv import Venv

from .repository import Repository


class InstalledRepository(Repository):

    @classmethod
    def load(cls, venv: Venv) -> 'InstalledRepository':
        """
        Load installed packages.

        For now, it uses the pip "freeze" command.
        """
        repo = cls()

        freeze_output = venv.run('pip', 'freeze')
        for line in freeze_output.split('\n'):
            if '==' in line:
                name, version = line.split('==')
                repo.add_package(Package(name, version, version))

        return repo



from .venv_command import VenvCommand


class RunCommand(VenvCommand):
    """
    Runs a command in the appropriate environment.

    run
        { args* : The command and arguments/options to run. }
    """

    def handle(self):
        args = self.argument('args')

        venv = self.venv

        return venv.exec(*args)

    def merge_application_definition(self, merge_args=True):
        if self._application is None \
                or (self._application_definition_merged
                    and (self._application_definition_merged_with_args or not merge_args)):
            return

        if merge_args:
            current_arguments = self._definition.get_arguments()
            self._definition.set_arguments(self._application.get_definition().get_arguments())
            self._definition.add_arguments(current_arguments)

        self._application_definition_merged = True
        if merge_args:
            self._application_definition_merged_with_args = True



from .operation import Operation


class Uninstall(Operation):

    def __init__(self, package, reason=None):
        super(Uninstall, self).__init__(reason)

        self._package = package

    @property
    def package(self):
        return self._package

    @property
    def job_type(self):
        return 'uninstall'

    def __str__(self):
        return 'Uninstalling {} ({})'.format(
            self.package.pretty_name,
            self.format_version(self._package)
        )

    def __repr__(self):
        return '<Uninstall {} ({})>'.format(
            self.package.pretty_name,
            self.format_version(self.package)
        )



from pathlib import Path

from poetry.utils.helpers import module_name


class Module:

    def __init__(self, name, directory='.'):
        self._name = module_name(name)

        # It must exist either as a .py file or a directory, but not both
        pkg_dir = Path(directory, self._name)
        py_file = Path(directory, self._name + '.py')
        if pkg_dir.is_dir() and py_file.is_file():
            raise ValueError("Both {} and {} exist".format(pkg_dir, py_file))
        elif pkg_dir.is_dir():
            self._path = pkg_dir
            self._is_package = True
        elif py_file.is_file():
            self._path = py_file
            self._is_package = False
        else:
            raise ValueError("No file/folder found for package {}".format(name))

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> Path:
        return self._path

    @property
    def file(self) -> Path:
        if self._is_package:
            return self._path / '__init__.py'
        else:
            return self._path

    def is_package(self) -> bool:
        return self._is_package



"""Generate and work with PEP 425 Compatibility Tags."""

import distutils.util
import sys
import warnings


def get_abbr_impl(venv):
    """Return abbreviated implementation name."""
    impl = venv.python_implementation

    if impl == 'PyPy':
        return 'pp'
    elif impl == 'Jython':
        return 'jy'
    elif impl == 'IronPython':
        return 'ip'
    elif impl == 'CPython':
        return 'cp'

    raise LookupError('Unknown Python implementation: ' + impl)


def get_impl_ver(venv):
    """Return implementation version."""
    impl_ver = venv.config_var("py_version_nodot")
    if not impl_ver or get_abbr_impl(venv) == 'pp':
        impl_ver = ''.join(map(str, get_impl_version_info(venv)))

    return impl_ver


def get_impl_version_info(venv):
    """Return sys.version_info-like tuple for use in decrementing the minor
    version."""
    if get_abbr_impl(venv) == 'pp':
        # as per https://github.com/pypa/pip/issues/2882
        return venv.version_info[:3]
    else:
        return venv.version_info[:2]


def get_flag(venv, var, fallback, expected=True, warn=True):
    """Use a fallback method for determining SOABI flags if the needed config
    var is unset or unavailable."""
    val = venv.config_var(var)
    if val is None:
        if warn:
            warnings.warn("Config variable '{0}' is unset, Python ABI tag may "
                          "be incorrect".format(var), RuntimeWarning, 2)
        return fallback()
    return val == expected


def get_abi_tag(venv):
    """Return the ABI tag based on SOABI (if available) or emulate SOABI
    (CPython 2, PyPy)."""
    soabi = venv.config_var('SOABI')
    impl = get_abbr_impl(venv)
    if not soabi and impl in ('cp', 'pp') and hasattr(sys, 'maxunicode'):
        d = ''
        m = ''
        u = ''
        if get_flag(venv,
                    'Py_DEBUG',
                    lambda: hasattr(sys, 'gettotalrefcount'),
                    warn=(impl == 'cp')):
            d = 'd'
        if get_flag(venv,
                    'WITH_PYMALLOC',
                    lambda: impl == 'cp',
                    warn=(impl == 'cp')):
            m = 'm'
        if get_flag(venv,
                    'Py_UNICODE_SIZE',
                    lambda: sys.maxunicode == 0x10ffff,
                    expected=4,
                    warn=(impl == 'cp' and
                          venv.version_info < (3, 3))) \
                and venv.version_info < (3, 3):
            u = 'u'
        abi = '%s%s%s%s%s' % (impl, get_impl_ver(venv), d, m, u)
    elif soabi and soabi.startswith('cpython-'):
        abi = 'cp' + soabi.split('-')[1]
    elif soabi:
        abi = soabi.replace('.', '_').replace('-', '_')
    else:
        abi = None
    return abi


def get_platform():
    """Return our platform name 'win32', 'linux_x86_64'"""
    # XXX remove distutils dependency
    result = distutils.util.get_platform().replace('.', '_').replace('-', '_')
    if result == "linux_x86_64" and sys.maxsize == 2147483647:
        # pip pull request #3497
        result = "linux_i686"
    return result


def get_supported(venv, versions=None, supplied_platform=None):
    """Return a list of supported tags for each version specified in
    `versions`.
    :param versions: a list of string versions, of the form ["33", "32"],
        or None. The first version will be assumed to support our ABI.
    """
    supported = []

    # Versions must be given with respect to the preference
    if versions is None:
        versions = []
        version_info = get_impl_version_info(venv)
        major = version_info[:-1]
        # Support all previous minor Python versions.
        for minor in range(version_info[-1], -1, -1):
            versions.append(''.join(map(str, major + (minor,))))

    impl = get_abbr_impl(venv)

    abis = []

    abi = get_abi_tag(venv)
    if abi:
        abis[0:0] = [abi]

    abi3s = set()
    import imp
    for suffix in imp.get_suffixes():
        if suffix[0].startswith('.abi'):
            abi3s.add(suffix[0].split('.', 2)[1])

    abis.extend(sorted(list(abi3s)))

    abis.append('none')

    platforms = []
    if supplied_platform:
        platforms.append(supplied_platform)
    platforms.append(get_platform())

    # Current version, current API (built specifically for our Python):
    for abi in abis:
        for arch in platforms:
            supported.append(('%s%s' % (impl, versions[0]), abi, arch))

    # abi3 modules compatible with older version of Python
    for version in versions[1:]:
        # abi3 was introduced in Python 3.2
        if version in ('31', '30'):
            break
        for abi in abi3s:   # empty set if not Python 3
            for arch in platforms:
                supported.append(("%s%s" % (impl, version), abi, arch))

    # No abi / arch, but requires our implementation:
    for i, version in enumerate(versions):
        supported.append(('%s%s' % (impl, version), 'none', 'any'))
        if i == 0:
            # Tagged specifically as being cross-version compatible
            # (with just the major version specified)
            supported.append(('%s%s' % (impl, versions[0][0]), 'none', 'any'))

    # Major Python version + platform; e.g. binaries not using the Python API
    supported.append(('py%s' % (versions[0][0]), 'none', arch))

    # No abi / arch, generic Python
    for i, version in enumerate(versions):
        supported.append(('py%s' % (version,), 'none', 'any'))
        if i == 0:
            supported.append(('py%s' % (version[0]), 'none', 'any'))

    return supported



from .base_constraint import BaseConstraint


class MultiConstraint(BaseConstraint):

    def __init__(self, constraints, conjunctive=True):
        self._constraints = tuple(constraints)
        self._conjunctive = conjunctive

    @property
    def constraints(self):
        return self._constraints

    def is_conjunctive(self):
        return self._conjunctive

    def is_disjunctive(self):
        return not self._conjunctive

    def matches(self, provider):
        if self.is_disjunctive():
            for constraint in self._constraints:
                if constraint.matches(provider):
                    return True

            return False

        for constraint in self._constraints:
            if not constraint.matches(provider):
                return False

        return True

    def __str__(self):
        constraints = []
        for constraint in self._constraints:
            constraints.append(str(constraint))

        return '{}'.format(
            (', ' if self._conjunctive else ' || ').join(constraints)
        )



from .action import Action


class Tag(Action):

    def __init__(self, tag):
        super(Tag, self).__init__()

        self._tag = tag

    @property
    def action_name(self):
        return 'tag'

    @property
    def tag(self):
        return self._tag

    def up(self, graph):
        pass

    def down(self, graph):
        pass



from .operation import Operation


class Install(Operation):

    def __init__(self, package, reason: str = None) -> None:
        super().__init__(reason)

        self._package = package

    @property
    def package(self):
        return self._package

    @property
    def job_type(self):
        return 'install'

    def __str__(self) -> str:
        return 'Installing {} ({})'.format(
            self.package.pretty_name,
            self.format_version(self.package)
        )

    def __repr__(self):
        return '<Install {} ({})>'.format(
            self.package.pretty_name,
            self.format_version(self.package)
        )



import os
import shutil

from functools import cmp_to_key
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict
from typing import List

from poetry.mixology import DependencyGraph
from poetry.mixology.conflict import Conflict
from poetry.mixology.contracts import SpecificationProvider

from poetry.packages import Dependency
from poetry.packages import Package
from poetry.packages import VCSDependency

from poetry.repositories import Pool

from poetry.semver import less_than

from poetry.utils.toml_file import TomlFile
from poetry.utils.venv import Venv

from poetry.vcs.git import Git


class Provider(SpecificationProvider):

    UNSAFE_PACKAGES = {'setuptools', 'distribute', 'pip'}

    def __init__(self, package: Package, pool: Pool):
        self._package = package
        self._pool = pool
        self._python_constraint = package.python_constraint

    @property
    def pool(self) -> Pool:
        return self._pool

    @property
    def name_for_explicit_dependency_source(self) -> str:
        return 'pyproject.toml'

    @property
    def name_for_locking_dependency_source(self) -> str:
        return 'pyproject.lock'

    def name_for(self, dependency: Dependency) -> str:
        """
        Returns the name for the given dependency.
        """
        return dependency.name

    def search_for(self, dependency: Dependency) -> List[Package]:
        """
        Search for the specifications that match the given dependency.

        The specifications in the returned list will be considered in reverse
        order, so the latest version ought to be last.
        """
        if dependency.is_vcs():
            return self.search_for_vcs(dependency)

        packages = self._pool.find_packages(
            dependency.name,
            dependency.constraint,
            extras=dependency.extras,
        )

        packages.sort(
            key=cmp_to_key(
                lambda x, y:
                0 if x.version == y.version
                else -1 * int(less_than(x.version, y.version) or -1)
            )
        )

        return packages

    def search_for_vcs(self, dependency: VCSDependency) -> List[Package]:
        """
        Search for the specifications that match the given VCS dependency.

        Basically, we clone the repository in a temporary directory
        and get the information we need by checking out the specified reference.
        """
        if dependency.vcs != 'git':
            raise ValueError(f'Unsupported VCS dependency {dependency.vcs}')

        tmp_dir = Path(mkdtemp(prefix=f'pypoetry-git-{dependency.name}'))

        try:
            git = Git()
            git.clone(dependency.source, tmp_dir)
            git.checkout(dependency.reference, tmp_dir)
            revision = git.rev_parse(
                dependency.reference, tmp_dir
            ).strip()

            if dependency.tag or dependency.rev:
                revision = dependency.reference

            poetry = TomlFile(tmp_dir / 'pyproject.toml')
            if poetry.exists():
                # If a pyproject.toml file exists
                # We use it to get the information we need
                info = poetry.read()

                name = info['package']['name']
                version = info['package']['version']
                package = Package(name, version, version)
                for req_name, req_constraint in info['dependencies'].items():
                    package.add_dependency(req_name, req_constraint)
            else:
                # We need to use setup.py here
                # to figure the information we need
                # We need to place ourselves in the proper
                # folder for it to work
                current_dir = os.getcwd()
                os.chdir(tmp_dir.as_posix())

                try:
                    venv = Venv.create()
                    output = venv.run(
                        'python', 'setup.py',
                        '--name', '--version'
                    )
                    output = output.split('\n')
                    name = output[-3]
                    version = output[-2]
                    package = Package(name, version, version)
                    # Figure out a way to get requirements
                except Exception:
                    raise
                finally:
                    os.chdir(current_dir)

            package.source_type = 'git'
            package.source_url = dependency.source
            package.source_reference = revision
        except Exception:
            raise
        finally:
            shutil.rmtree(tmp_dir.as_posix())

        return [package]

    def dependencies_for(self, package: Package):
        if package.source_type == 'git':
            # Information should already be set
            pass
        else:
            package = self._pool.package(package.name, package.version)

        return [
            r for r in package.requires
            if not r.is_optional()
            and r.name not in self.UNSAFE_PACKAGES
        ]

    def is_requirement_satisfied_by(self,
                                    requirement: Dependency,
                                    activated: DependencyGraph,
                                    package: Package) -> bool:
        """
        Determines whether the given requirement is satisfied by the given
        spec, in the context of the current activated dependency graph.
        """
        if isinstance(requirement, Package):
            return requirement == package

        if not requirement.accepts(package):
            return False

        if package.is_prerelease() and not requirement.allows_prereleases():
            vertex = activated.vertex_named(package.name)

            if not any([r.allows_prereleases() for r in vertex.requirements]):
                return False

        return (
            self._package.python_constraint.matches(package.python_constraint)
            and self._package.platform_constraint.matches(package.platform_constraint)
        )

    def sort_dependencies(self,
                          dependencies: List[Dependency],
                          activated: DependencyGraph,
                          conflicts: Dict[str, List[Conflict]]):
        return sorted(dependencies, key=lambda d: [
            0 if activated.vertex_named(d.name).payload else 1,
            0 if d.allows_prereleases() else 1,
            0 if d.name in conflicts else 1,
            0 if activated.vertex_named(d.name).payload else len(self.search_for(d))
        ])



from typing import List

from poetry.mixology import Resolver
from poetry.mixology.dependency_graph import DependencyGraph
from poetry.mixology.exceptions import ResolverError

from poetry.semver.version_parser import VersionParser

from .exceptions import SolverProblemError
from .operations import Install
from .operations import Uninstall
from .operations import Update
from .operations.operation import Operation

from .provider import Provider
from .ui import UI


class Solver:

    def __init__(self, package, pool, locked, io):
        self._package = package
        self._pool = pool
        self._locked = locked
        self._io = io

    def solve(self, requested, fixed=None) -> List[Operation]:
        resolver = Resolver(Provider(self._package, self._pool), UI(self._io))

        base = None
        if fixed is not None:
            base = DependencyGraph()
            for fixed_req in fixed:
                base.add_vertex(fixed_req.name, fixed_req, True)

        try:
            graph = resolver.resolve(requested, base=base)
        except ResolverError as e:
            raise SolverProblemError(e)

        packages = [v.payload for v in graph.vertices.values()]

        # Setting info
        for vertex in graph.vertices.values():
            tags = self._get_tags_for_vertex(vertex, requested)
            if 'main' in tags['category']:
                vertex.payload.category = 'main'
            else:
                vertex.payload.category = 'dev'

            if not tags['optional']:
                vertex.payload.optional = False
            else:
                vertex.payload.optional = True

            # Finding the less restrictive requirements
            requirements = {}
            parser = VersionParser()
            for req_name, reqs in tags['requirements'].items():
                for req in reqs:
                    if req_name == 'python':
                        if 'python' not in requirements:
                            requirements['python'] = req
                            continue

                        previous = parser.parse_constraints(requirements['python'])
                        current = parser.parse_constraints(req)

                        if current.matches(previous):
                            requirements['python'] = req

                    if req_name == 'platform':
                        if 'platform' not in requirements:
                            requirements['platform'] = req
                            continue

            vertex.payload.requirements = requirements

        operations = []
        for package in packages:
            installed = False
            for pkg in self._locked.packages:
                if package.name == pkg.name:
                    installed = True
                    # Checking version
                    if package.version != pkg.version:
                        operations.append(Update(pkg, package))

                    break

            if not installed:
                operations.append(Install(package))

        # Checking for removals
        for pkg in self._locked.packages:
            remove = True
            for package in packages:
                if pkg.name == package.name:
                    remove = False
                    break

            if remove:
                operations.append(Uninstall(pkg))

        return list(reversed(operations))

    def _get_tags_for_vertex(self, vertex, requested):
        tags = {
            'category': [],
            'optional': True,
            'requirements': {
                'python': [],
                'platform': []
            }
        }

        if not vertex.incoming_edges:
            # Original dependency
            for req in requested:
                if req.name == vertex.name:
                    tags['category'].append(req.category)
                    if not req.is_optional():
                        tags['optional'] = False

                    if req.python_versions != '*':
                        tags['requirements']['python'].append(str(req.python_constraint))

                    if req.platform != '*':
                        tags['requirements']['platform'].append(str(req.platform_constraint))

                    break
        else:
            for edge in vertex.incoming_edges:
                for req in edge.origin.payload.requires:
                    if req.name == vertex.payload.name:
                        if req.python_versions != '*':
                            tags['requirements']['python'].append(req.python_versions)

                        if req.platform != '*':
                            tags['requirements']['platform'].append(req.platform)

                sub_tags = self._get_tags_for_vertex(edge.origin, requested)

                tags['category'] += sub_tags['category']
                tags['optional'] = tags['optional'] and sub_tags['optional']
                requirements = sub_tags['requirements']
                tags['requirements']['python'] += requirements.get('python', [])
                tags['requirements']['platform'] += requirements.get('platform', [])

        return tags



"""
PEP-517 compliant buildsystem API
"""
import logging
from pathlib import Path

from poetry import Poetry
from poetry.io import NullIO

from .builders import SdistBuilder
from .builders import WheelBuilder

log = logging.getLogger(__name__)

# PEP 517 specifies that the CWD will always be the source tree
poetry = Poetry.create('.')


def get_requires_for_build_wheel(config_settings=None):
    """
    Returns a list of requirements for building, as strings
    """
    main, extras = SdistBuilder.convert_dependencies(poetry.package.requires)

    return main + extras


# For now, we require all dependencies to build either a wheel or an sdist.
get_requires_for_build_sdist = get_requires_for_build_wheel


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """Builds a wheel, places it in wheel_directory"""
    info = WheelBuilder.make_in(poetry, NullIO(), Path(wheel_directory))

    return info.file.name


def build_sdist(sdist_directory, config_settings=None):
    """Builds an sdist, places it in sdist_directory"""
    path = SdistBuilder(poetry, NullIO()).build(Path(sdist_directory))

    return path.name



from cleo import Command as BaseCommand
from cleo.inputs import ListInput

from poetry.poetry import Poetry

from ..styles.poetry import PoetryStyle


class Command(BaseCommand):

    @property
    def poetry(self) -> Poetry:
        return self.get_application().poetry

    def reset_poetry(self) -> None:
        self.get_application().reset_poetry()

    def run(self, i, o) -> int:
        """
        Initialize command.
        """
        self.input = i
        self.output = PoetryStyle(i, o)

        return super(BaseCommand, self).run(i, o)



import re

from .constraints.constraint import Constraint
from .constraints.empty_constraint import EmptyConstraint
from .constraints.multi_constraint import MultiConstraint
from .constraints.wildcard_constraint import WilcardConstraint
from .helpers import normalize_version, _expand_stability


class VersionParser:

    _modifier_regex = (
        '[._-]?'
        '(?:(stable|beta|b|RC|alpha|a|patch|pl|p)((?:[.-]?\d+)*)?)?'
        '([.-]?dev)?'
    )

    _stabilities = [
        'stable', 'RC', 'beta', 'alpha', 'dev'
    ]

    def parse_constraints(self, constraints: str):
        """
        Parses a constraint string into
        MultiConstraint and/or Constraint objects.
        """
        pretty_constraint = constraints

        m = re.match(
            '(?i)([^,\s]*?)@({})$'.format('|'.join(self._stabilities)),
            constraints
        )
        if m:
            constraints = m.group(1)
            if not constraints:
                constraints = '*'

        or_constraints = re.split('\s*\|\|?\s*', constraints.strip())
        or_groups = []
        for constraints in or_constraints:
            and_constraints = re.split(
                '(?<!^)(?<![=>< ,]) *(?<!-)[, ](?!-) *(?!,|$)',
                constraints
            )
            if len(and_constraints) > 1:
                constraint_objects = []
                for constraint in and_constraints:
                    for parsed_constraint in self._parse_constraint(constraint):
                        constraint_objects.append(parsed_constraint)
            else:
                constraint_objects = self._parse_constraint(and_constraints[0])

            if len(constraint_objects) == 1:
                constraint = constraint_objects[0]
            else:
                constraint = MultiConstraint(constraint_objects)

            or_groups.append(constraint)

        if len(or_groups) == 1:
            constraint = or_groups[0]
        elif len(or_groups) == 2:
            # parse the two OR groups and if they are contiguous we collapse
            # them into one constraint
            a = str(or_groups[0])
            b = str(or_groups[1])
            pos_a = a.find('<', 4)
            pos_b = a.find('<', 4)
            if (
                isinstance(or_groups[0], MultiConstraint)
                and isinstance(or_groups[1], MultiConstraint)
                and len(or_groups[0].constraints)
                and len(or_groups[1].constraints)
                and a[:3] == '>=' and pos_a != -1
                and b[:3] == '>=' and pos_b != -1
                and a[pos_a + 2:-1] == b[4:pos_b - 5]
            ):
                constraint = MultiConstraint(
                    Constraint('>=', a[4:pos_a - 5]),
                    Constraint('<', b[pos_b + 2:-1])
                )
            else:
                constraint = MultiConstraint(or_groups, False)
        else:
            constraint = MultiConstraint(or_groups, False)

        constraint.pretty_string = pretty_constraint

        return constraint

    def _parse_constraint(self, constraint):
        m = re.match('(?i)^v?[xX*](\.[xX*])*$', constraint)
        if m:
            return EmptyConstraint(),

        version_regex = (
            'v?(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:\.(\d+))?{}(?:\+[^\s]+)?'
        ).format(self._modifier_regex)

        # Tilde range
        #
        # Like wildcard constraints, unsuffixed tilde constraints
        # say that they must be greater than the previous version,
        # to ensure that unstable instances of the current version are allowed.
        # However, if a stability suffix is added to the constraint,
        # then a >= match on the current version is used instead.
        m = re.match('(?i)^~=?{}$'.format(version_regex), constraint)
        if m:
            # Work out which position in the version we are operating at
            if m.group(4):
                position = 3
            elif m.group(3):
                position = 2
            elif m.group(2):
                position = 2
            else:
                position = 0

            # Calculate the stability suffix
            stability_suffix = ''
            if m.group(5):
                stability_suffix += '-{}{}'.format(
                    _expand_stability(m.group(5)),
                    '.' + m.group(6) if m.group(6) else ''
                )

            low_version = self._manipulate_version_string(
                m.groups(), position, 0
            ) + stability_suffix
            lower_bound = Constraint('>=', low_version)

            # For upper bound,
            # we increment the position of one more significance,
            # but high_position = 0 would be illegal
            high_position = max(0, position - 1)
            high_version = self._manipulate_version_string(
                m.groups(), high_position, 1
            )
            upper_bound = Constraint('<', high_version)

            return lower_bound, upper_bound

        # Caret range
        #
        # Allows changes that do not modify
        # the left-most non-zero digit in the [major, minor, patch] tuple.
        # In other words, this allows:
        #     - patch and minor updates for versions 1.0.0 and above,
        #     - patch updates for versions 0.X >=0.1.0,
        #     - and no updates for versions 0.0.X
        m = re.match('^\^{}($)'.format(version_regex), constraint)
        if m:
            if m.group(1) != '0' or not m.group(2):
                position = 0
            elif m.group(2) != '0' or not m.group(3):
                position = 1
            else:
                position = 2

            low_version = normalize_version(constraint[1:])
            lower_bound = Constraint('>=', low_version)

            # For upper bound,
            # we increment the position of one more significance,
            # but high_position = 0 would be illegal
            high_version = self._manipulate_version_string(
                m.groups(), position, 1
            )
            upper_bound = Constraint('<', high_version)

            return lower_bound, upper_bound

        # X range
        #
        # Any of X, x, or * may be used to "stand in"
        # for one of the numeric values in the [major, minor, patch] tuple.
        # A partial version range is treated as an X-Range,
        # so the special character is in fact optional.
        m = re.match(
            '^(!=|==)?v?(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:\.[xX*])+$',
            constraint
        )
        if m:
            # We just leave it as is
            return WilcardConstraint(constraint),

        # Basic Comparators
        m = re.match('^(<>|!=|>=?|<=?|==?)?\s*(.*)', constraint)
        if m:
            try:
                version = normalize_version(m.group(2))

                return Constraint(m.group(1) or '=', version),
            except ValueError:
                pass

        raise ValueError(
            'Could not parse version constraint: {}'.format(constraint)
        )

    def _manipulate_version_string(self, matches, position,
                                   increment=0, pad='0'):
        """
        Increment, decrement, or simply pad a version number.
        """
        matches = [matches[i]
                   if i < len(matches) - 1 and matches[i] is not None else '0'
                   for i in range(4)]
        for i in range(3, -1, -1):
            if i > position:
                matches[i] = pad
            elif i == position and increment:
                matches[i] = int(matches[i]) + increment
                # If $matches[i] was 0, carry the decrement
                if matches[i] < 0:
                    matches[i] = pad
                    position -= 1

                    # Return null on a carry overflow
                    if i == 1:
                        return

        return '{}.{}.{}.{}'.format(matches[0], matches[1],
                                    matches[2], matches[3])



from .operation import Operation


class Update(Operation):

    def __init__(self, initial, target, reason=None):
        self._initial_package = initial
        self._target_package = target
        
        super(Update, self).__init__(reason)

    @property
    def initial_package(self):
        return self._initial_package

    @property
    def target_package(self):
        return self._target_package

    @property
    def job_type(self):
        return 'update'

    def __str__(self):
        return (
            'Updating {} ({}) to {} ({})'.format(
                self.initial_package.pretty_name,
                self.format_version(self.initial_package),
                self.target_package.pretty_name,
                self.format_version(self.target_package)
            )
        )

    def __repr__(self):
        return (
            '<Update {} ({}) to {} ({})>'.format(
                self.initial_package.pretty_name,
                self.format_version(self.initial_package),
                self.target_package.pretty_name,
                self.format_version(self.target_package)
            )
        )



# -*- coding: utf-8 -*-


class Operation:

    def __init__(self, reason: str = None) -> None:
        self._reason = reason

        self._skipped = False
        self._skip_reason = None

    @property
    def job_type(self) -> str:
        raise NotImplementedError

    @property
    def reason(self) -> str:
        return self._reason

    @property
    def skipped(self) -> bool:
        return self._skipped

    @property
    def skip_reason(self):
        return self._skip_reason

    def format_version(self, package) -> str:
        return package.full_pretty_version

    def skip(self, reason: str) -> None:
        self._skipped = True
        self._skip_reason = reason



import os

from cleo import Application as BaseApplication
from cleo.inputs import ArgvInput
from cleo.outputs import ConsoleOutput

from poetry.io.raw_argv_input import RawArgvInput
from poetry.poetry import Poetry

from .commands import AboutCommand
from .commands import AddCommand
from .commands import BuildCommand
from .commands import CheckCommand
from .commands import ConfigCommand
from .commands import InstallCommand
from .commands import LockCommand
from .commands import NewCommand
from .commands import PublishCommand
from .commands import RemoveCommand
from .commands import RunCommand
from .commands import ShowCommand
from .commands import UpdateCommand

from .commands.debug import DebugResolveCommand


class Application(BaseApplication):

    def __init__(self):
        super().__init__('Poetry', Poetry.VERSION)

        self._poetry = None
        self._skip_io_configuration = False

    @property
    def poetry(self) -> Poetry:
        if self._poetry is not None:
            return self._poetry

        self._poetry = Poetry.create(os.getcwd())

        return self._poetry

    def reset_poetry(self) -> None:
        self._poetry = None

    def run(self, i=None, o=None) -> int:
        if i is None:
            i = ArgvInput()

        if o is None:
            o = ConsoleOutput()

        name = i.get_first_argument()
        if name == 'run':
            self._skip_io_configuration = True
            i = RawArgvInput()

        return super().run(i, o)

    def do_run(self, i, o):
        name = self.get_command_name(i)

        if name != 'run':
            return super().do_run(i, o)

        command = self.find(name)

        self._running_command = command
        status_code = command.run(i, o)
        self._running_command = None

        return status_code

    def configure_io(self, i, o):
        if self._skip_io_configuration:
            return

        super().configure_io(i, o)

    def get_default_commands(self) -> list:
        commands = super(Application, self).get_default_commands()

        commands += [
            AboutCommand(),
            AddCommand(),
            BuildCommand(),
            CheckCommand(),
            ConfigCommand(),
            InstallCommand(),
            LockCommand(),
            NewCommand(),
            PublishCommand(),
            RemoveCommand(),
            RunCommand(),
            ShowCommand(),
            UpdateCommand(),
        ]

        # Debug commands
        commands += [
            DebugResolveCommand(),
        ]

        return commands



from poetry.semver.constraints import MultiConstraint
from poetry.semver.version_parser import VersionParser


PYTHON_VERSION = [
    '2.7.*',
    '3.0.*', '3.1.*', '3.2.*', '3.3.*', '3.4.*',
    '3.5.*', '3.6.*', '3.7.*', '3.8.*',
]


def format_python_constraint(constraint):
    """
    This helper will help in transforming
    disjunctive constraint into proper constraint.
    """
    if not isinstance(constraint, MultiConstraint):
        return str(constraint)

    has_disjunctive = False
    for c in constraint.constraints:
        if isinstance(c, MultiConstraint) and c.is_disjunctive():
            has_disjunctive = True
            break

    parser = VersionParser()
    formatted = []
    accepted = []
    if not constraint.is_disjunctive() and not has_disjunctive:
        return str(constraint)

    for version in PYTHON_VERSION:
        version_constraint = parser.parse_constraints(version)
        matches = constraint.matches(version_constraint)
        if not matches:
            formatted.append('!=' + version)
        else:
            accepted.append(version)

    # Checking lower bound
    low = accepted[0]

    formatted.insert(0, '>=' + '.'.join(low.split('.')[:2]))

    return ', '.join(formatted)



import os
import tempfile

from subprocess import CalledProcessError

from poetry.utils.venv import Venv

from .base_installer import BaseInstaller


class PipInstaller(BaseInstaller):

    def __init__(self, venv: Venv, io):
        self._venv = venv
        self._io = io

    def install(self, package, update=False):
        args = ['install', '--no-deps']

        if package.source_type == 'legacy' and package.source_url:
            args += ['--index-url', package.source_url]

        if update:
            args.append('-U')

        if package.hashes and not package.source_type:
            # Format as a requirements.txt
            # We need to create a requirements.txt file
            # for each package in order to check hashes.
            # This is far from optimal but we do not have any
            # other choice since this is the only way for pip
            # to verify hashes.
            req = self.create_temporary_requirement(package)
            args += ['-r', req]

            try:
                self.run(*args)
            finally:
                os.unlink(req)
        else:
            args.append(self.requirement(package))

            self.run(*args)

    def update(self, _, target):
        self.install(target, update=True)

    def remove(self, package):
        try:
            self.run('uninstall', package.name, '-y')
        except CalledProcessError as e:
            if 'not installed' in str(e):
                return

            raise

    def run(self, *args, **kwargs) -> str:
        return self._venv.run('pip', *args, **kwargs)

    def requirement(self, package, formatted=False) -> str:
        if formatted and not package.source_type == 'git':
            req = f'{package.name}=={package.version}'
            for h in package.hashes:
                req += f' --hash sha256:{h}'

            req += '\n'

            return req

        if package.source_type == 'git':
            return f'git+{package.source_url}@{package.source_reference}' \
                   f'#egg={package.name}'

        return f'{package.name}=={package.version}'

    def create_temporary_requirement(self, package):
        fd, name = tempfile.mkstemp('reqs.txt', f'{package.name}-{package.version}')

        with open(fd, 'w') as f:
            f.write(self.requirement(package, formatted=True))

        return name



from . import factory, traversal
from .common import Element, ContainerElement
from .factory import create_element
from .metadata import NewlineElement
from .errors import InvalidElementError


class ArrayElement(ContainerElement, traversal.TraversalMixin, list):
    """
    A sequence-like container element containing other atomic elements or other containers.

    Implements list-like interface.

    Assumes input sub_elements are correct for an array element.

    Raises an InvalidElementError if contains heterogeneous values.
    """

    def __init__(self, sub_elements):
        super(ArrayElement, self).__init__(sub_elements)
        self._check_homogeneity()

    def _check_homogeneity(self):
        if len(set(type(v) for v in self.primitive_value)) > 1:
            raise InvalidElementError('Array should be homogeneous')

    def __len__(self):
        return len(tuple(self._enumerate_non_metadata_sub_elements()))

    def __getitem__(self, i):
        """
        Returns the ith entry, which can be a primitive value, a seq-lie, or a dict-like object.
        """
        return self._find_value(i)[1].value

    def __setitem__(self, i, value):
        value_i, _ = self._find_value(i)
        new_element = value if isinstance(value, Element) else factory.create_element(value)
        self._sub_elements = self.sub_elements[:value_i] + [new_element] + self.sub_elements[value_i+1:]

    @property
    def value(self):
        return self     # self is a sequence-like value

    @property
    def primitive_value(self):
        """
        Returns a primitive Python value without any formatting or markup metadata.
        """
        return list(
            self[i].primitive_value if hasattr(self[i], 'primitive_value')
            else self[i]
            for i in range(len(self)))

    def __str__(self):
        return "{}".format(self.primitive_value)

    def __repr__(self):
        return "Array{}".format(str(self))

    def append(self, v):
        new_entry = [create_element(v)]

        if self:    # If not empty, we need a comma and whitespace prefix!
            new_entry = [
                factory.create_operator_element(','),
                factory.create_whitespace_element(),
            ] + new_entry

        insertion_index = self._find_closing_square_bracket()
        self._sub_elements = self._sub_elements[:insertion_index] + new_entry + \
                             self._sub_elements[insertion_index:]

    def _find_value(self, i):
        """
        Returns (value_index, value) of ith value in this sequence.

        Raises IndexError if not found.
        """
        return tuple(self._enumerate_non_metadata_sub_elements())[i]

    def __delitem__(self, i):
        value_i, value = self._find_value(i)

        begin, end = value_i, value_i+1

        # Rules:
        #   1. begin should be index to the preceding comma to the value
        #   2. end should be index to the following comma, or the closing bracket
        #   3. If no preceding comma found but following comma found then end should be the index of the following value

        preceding_comma = self._find_preceding_comma(value_i)
        found_preceding_comma = preceding_comma >= 0
        if found_preceding_comma:
            begin = preceding_comma

        following_comma = self._find_following_comma(value_i)
        if following_comma >= 0:
            if not found_preceding_comma:
                end = self._find_following_non_metadata(following_comma)
            else:
                end = following_comma
        else:
            end = self._find_following_closing_square_bracket(0)

        self._sub_elements = self.sub_elements[:begin] + self._sub_elements[end:]

    @property
    def is_multiline(self):
        return any(isinstance(e, (NewlineElement)) for e in self.elements)

    def turn_into_multiline(self):
        """
        Turns this array into a multi-line array with each element lying on its own line.
        """
        if self.is_multiline:
            return

        i = self._find_following_comma(-1)

        def next_entry_i():
            return self._find_following_non_metadata(i)

        def next_newline_i():
            return self._find_following_newline(i)

        def next_closing_bracket_i():
            return self._find_following_closing_square_bracket(i)

        def next_comma_i():
            return self._find_following_comma(i)

        while i < len(self.elements)-1:
            if next_newline_i() < next_entry_i():
                self.elements.insert(i+1, factory.create_newline_element())
                if float('-inf') < next_comma_i() < next_closing_bracket_i():
                    i = next_comma_i()
                else:
                    i = next_closing_bracket_i()



import posixpath
import urllib.parse
import urllib.request
import re

from .utils import path_to_url
from .utils import splitext


class Link:

    def __init__(self, url, comes_from=None, requires_python=None):
        """
        Object representing a parsed link from https://pypi.python.org/simple/*

        url:
            url of the resource pointed to (href of the link)
        comes_from:
            instance of HTMLPage where the link was found, or string.
        requires_python:
            String containing the `Requires-Python` metadata field, specified
            in PEP 345. This may be specified by a data-requires-python
            attribute in the HTML link tag, as described in PEP 503.
        """

        # url can be a UNC windows share
        if url.startswith('\\\\'):
            url = path_to_url(url)

        self.url = url
        self.comes_from = comes_from
        self.requires_python = requires_python if requires_python else None

    def __str__(self):
        if self.requires_python:
            rp = ' (requires-python:%s)' % self.requires_python
        else:
            rp = ''
        if self.comes_from:
            return '%s (from %s)%s' % (self.url, self.comes_from, rp)
        else:
            return str(self.url)

    def __repr__(self):
        return '<Link %s>' % self

    def __eq__(self, other):
        if not isinstance(other, Link):
            return NotImplemented
        return self.url == other.url

    def __ne__(self, other):
        if not isinstance(other, Link):
            return NotImplemented
        return self.url != other.url

    def __lt__(self, other):
        if not isinstance(other, Link):
            return NotImplemented
        return self.url < other.url

    def __le__(self, other):
        if not isinstance(other, Link):
            return NotImplemented
        return self.url <= other.url

    def __gt__(self, other):
        if not isinstance(other, Link):
            return NotImplemented
        return self.url > other.url

    def __ge__(self, other):
        if not isinstance(other, Link):
            return NotImplemented
        return self.url >= other.url

    def __hash__(self):
        return hash(self.url)

    @property
    def filename(self):
        _, netloc, path, _, _ = urllib.parse.urlsplit(self.url)
        name = posixpath.basename(path.rstrip('/')) or netloc
        name = urllib.parse.unquote(name)
        assert name, ('URL %r produced no filename' % self.url)
        return name

    @property
    def scheme(self):
        return urllib.parse.urlsplit(self.url)[0]

    @property
    def netloc(self):
        return urllib.parse.urlsplit(self.url)[1]

    @property
    def path(self):
        return urllib.parse.unquote(urllib.parse.urlsplit(self.url)[2])

    def splitext(self):
        return splitext(posixpath.basename(self.path.rstrip('/')))

    @property
    def ext(self):
        return self.splitext()[1]

    @property
    def url_without_fragment(self):
        scheme, netloc, path, query, fragment = urllib.parse.urlsplit(self.url)
        return urllib.parse.urlunsplit((scheme, netloc, path, query, None))

    _egg_fragment_re = re.compile(r'[#&]egg=([^&]*)')

    @property
    def egg_fragment(self):
        match = self._egg_fragment_re.search(self.url)
        if not match:
            return None
        return match.group(1)

    _subdirectory_fragment_re = re.compile(r'[#&]subdirectory=([^&]*)')

    @property
    def subdirectory_fragment(self):
        match = self._subdirectory_fragment_re.search(self.url)
        if not match:
            return None
        return match.group(1)

    _hash_re = re.compile(
        r'(sha1|sha224|sha384|sha256|sha512|md5)=([a-f0-9]+)'
    )

    @property
    def hash(self):
        match = self._hash_re.search(self.url)
        if match:
            return match.group(2)
        return None

    @property
    def hash_name(self):
        match = self._hash_re.search(self.url)
        if match:
            return match.group(1)
        return None

    @property
    def show_url(self):
        return posixpath.basename(self.url.split('#', 1)[0].split('?', 1)[0])

    @property
    def is_wheel(self):
        return self.ext == '.whl'

    @property
    def is_artifact(self):
        """
        Determines if this points to an actual artifact (e.g. a tarball) or if
        it points to an "abstract" thing like a path or a VCS location.
        """
        if self.scheme in ['ssh', 'git', 'hg', 'bzr', 'sftp', 'svn']:
            return False

        return True



import hashlib
import io
import re

import requests
import toml

from pathlib import Path

from requests import adapters
from requests.exceptions import HTTPError
from requests.packages.urllib3 import util
from requests_toolbelt import user_agent
from requests_toolbelt.multipart import (
    MultipartEncoder, MultipartEncoderMonitor
)

from poetry import __version__
from poetry.locations import CONFIG_DIR

from ..metadata import Metadata


wheel_file_re = re.compile(
    r"""^(?P<namever>(?P<name>.+?)(-(?P<ver>\d.+?))?)
        ((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)
        \.whl|\.dist-info)$""",
    re.VERBOSE
)

KEYWORDS_TO_NOT_FLATTEN = {'gpg_signature', 'content'}


class Publisher:
    """
    Registers and publishes packages to remote repositories.
    """

    def __init__(self, poetry, io):
        self._poetry = poetry
        self._package = poetry.package
        self._io = io

    def publish(self, repository_name):
        if repository_name:
            self._io.writeln(
                f'Publishing <info>{self._package.pretty_name}</info> '
                f'(<comment>{self._package.pretty_version}</comment>) '
                f'to <fg=cyan>{repository_name}</>'
            )
        else:
            self._io.writeln(
                f'Publishing <info>{self._package.pretty_name}</info> '
                f'(<comment>{self._package.pretty_version}</comment>) '
                f'to <fg=cyan>PyPI</>'
            )

        if not repository_name:
            url = 'https://upload.pypi.org/legacy/'
            repository_name = 'pypi'
        else:
            # Retrieving config information
            config_file = Path(CONFIG_DIR) / 'config.toml'

            if not config_file.exists():
                raise RuntimeError(
                    'Config file does not exist. '
                    'Unable to get repository information'
                )

            with config_file.open() as f:
                config = toml.loads(f.read())

            if (
                'repositories' not in config
                or repository_name not in config['repositories']
            ):
                raise RuntimeError(
                    f'Repository {repository_name} is not defined'
                )

            url = config['repositories'][repository_name]['url']

        username = None
        password = None
        auth_file = Path(CONFIG_DIR) / 'auth.toml'
        if auth_file.exists():
            with auth_file.open() as f:
                auth_config = toml.loads(f.read())

            if 'http-basic' in auth_config and repository_name in auth_config['http-basic']:
                config = auth_config['http-basic'][repository_name]

                username = config.get('username')
                password = config.get('password')

        # Requesting missing credentials
        if not username:
            username = self._io.ask('Username:')

        if not password:
            password = self._io.ask_hidden('Password:')

        session = requests.session()
        session.auth = (username, password)
        session.headers['User-Agent'] = self._make_user_agent_string()
        for scheme in ('http://', 'https://'):
            session.mount(scheme, self._make_adapter_with_retries())

        # TODO: handle certificates

        try:
            self.upload(session, url)
        finally:
            session.close()

    def register(self, session, url):
        """
        Register a package to a repository.
        """
        dist = self._poetry.file.parent / 'dist'
        file = dist / f'{self._package.name}-{self._package.version}.tar.gz'

        if not file.exists():
            raise RuntimeError(
                '"{0}" does not exist on the file system.'.format(file.name)
            )

        data = self.post_data(file)
        data.update({
            ":action": "submit",
            "protocol_version": "1",
        })

        data_to_send = self._convert_data_to_list_of_tuples(data)
        encoder = MultipartEncoder(data_to_send)
        resp = session.post(
            url,
            data=encoder,
            allow_redirects=False,
            headers={'Content-Type': encoder.content_type},
        )

        return resp

    def upload(self, session, url):
        """
        Upload packages for the current project.
        """
        try:
            self._upload(session, url)
        except HTTPError as e:
            if (
                e.response.status_code not in (403, 400)
                or e.response.status_code == 400 and 'was ever registered' not in e.response.text
            ):
                raise

            # It may be the first time we publish the package
            # We'll try to register it and go from there
            try:
                self.register(session, url)
            except HTTPError:
                raise

    def post_data(self, file):
        meta = Metadata.from_package(self._package)

        file_type = self._get_type(file)

        blake2_256_hash = hashlib.blake2b(digest_size=256 // 8)
        md5_hash = hashlib.md5()
        sha2_hash = hashlib.sha256()
        with file.open('rb') as fp:
            for content in iter(lambda: fp.read(io.DEFAULT_BUFFER_SIZE), b''):
                md5_hash.update(content)
                sha2_hash.update(content)
                blake2_256_hash.update(content)

        md5_digest = md5_hash.hexdigest()
        sha2_digest = sha2_hash.hexdigest()
        blake2_256_digest = blake2_256_hash.hexdigest()

        if file_type == 'bdist_wheel':
            wheel_info = wheel_file_re.match(file.name)
            py_version = wheel_info.group("pyver")
        else:
            py_version = None

        data = {
            # identify release
            "name": meta.name,
            "version": meta.version,

            # file content
            "filetype": file_type,
            "pyversion": py_version,

            # additional meta-data
            "metadata_version": meta.metadata_version,
            "summary": meta.summary,
            "home_page": meta.home_page,
            "author": meta.author,
            "author_email": meta.author_email,
            "maintainer": meta.maintainer,
            "maintainer_email": meta.maintainer_email,
            "license": meta.license,
            "description": meta.description,
            "keywords": meta.keywords,
            "platform": meta.platforms,
            "classifiers": meta.classifiers,
            "download_url": meta.download_url,
            "supported_platform": meta.supported_platforms,
            "comment": None,
            "md5_digest": md5_digest,
            "sha256_digest": sha2_digest,
            "blake2_256_digest": blake2_256_digest,

            # PEP 314
            "provides": meta.provides,
            "requires": meta.requires,
            "obsoletes": meta.obsoletes,

            # Metadata 1.2
            "project_urls": meta.project_urls,
            "provides_dist": meta.provides_dist,
            "obsoletes_dist": meta.obsoletes_dist,
            "requires_dist": meta.requires_dist,
            "requires_external": meta.requires_external,
            "requires_python": meta.requires_python,
        }

        # Metadata 2.1
        if meta.description_content_type:
            data['description_content_type'] = meta.description_content_type

        # TODO: Provides extra

        return data

    def _upload(self, session, url):
        dist = self._poetry.file.parent / 'dist'
        packages = dist.glob(f'{self._package.name}-{self._package.version}*')
        files = (
            i for i in packages if (
                i.match(f'{self._package.name}-{self._package.version}-*.whl')
                or
                i.match(f'{self._package.name}-{self._package.version}.tar.gz')
            )
        )

        for file in files:
            # TODO: Check existence

            resp = self._upload_file(session, url, file)

            # Bug 92. If we get a redirect we should abort because something seems
            # funky. The behaviour is not well defined and redirects being issued
            # by PyPI should never happen in reality. This should catch malicious
            # redirects as well.
            if resp.is_redirect:
                raise RuntimeError(
                    ('"{0}" attempted to redirect to "{1}" during upload.'
                     ' Aborting...').format(url, resp.headers["location"]))

            resp.raise_for_status()

    def _upload_file(self, session, url, file):
        data = self.post_data(file)
        data.update({
            # action
            ":action": "file_upload",
            "protocol_version": "1",
        })

        data_to_send = self._convert_data_to_list_of_tuples(data)

        with file.open('rb') as fp:
            data_to_send.append((
                "content",
                (file.name, fp, "application/octet-stream"),
            ))
            encoder = MultipartEncoder(data_to_send)
            bar = self._io.create_progress_bar(encoder.len)
            bar.set_format(
                " - Uploading <info>{0}</> <comment>%percent%%</>".format(
                    file.name
                )
            )
            monitor = MultipartEncoderMonitor(
                encoder, lambda monitor: bar.set_progress(monitor.bytes_read)
            )

            bar.start()

            resp = session.post(
                url,
                data=monitor,
                allow_redirects=False,
                headers={'Content-Type': monitor.content_type}
            )

            if resp.ok:
                bar.finish()

                self._io.writeln('')
            else:
                self._io.overwrite('')

        return resp

    def _convert_data_to_list_of_tuples(self, data):
        data_to_send = []
        for key, value in data.items():
            if (key in KEYWORDS_TO_NOT_FLATTEN or
                    not isinstance(value, (list, tuple))):
                data_to_send.append((key, value))
            else:
                for item in value:
                    data_to_send.append((key, item))
        return data_to_send

    def _get_type(self, file):
        exts = file.suffixes
        if exts[-1] == '.whl':
            return 'bdist_wheel'
        elif len(exts) >= 2 and ''.join(exts[-2:]) == '.tar.gz':
            return 'sdist'

        raise ValueError(
            f'Unknown distribution format {"".join(exts)}'
        )

    @staticmethod
    def _make_adapter_with_retries():
        retry = util.Retry(
            connect=5,
            total=10,
            method_whitelist=['GET'],
            status_forcelist=[500, 501, 502, 503],
        )
        return adapters.HTTPAdapter(max_retries=retry)

    @staticmethod
    def _make_user_agent_string():
        return user_agent(
            'twine', __version__,
        )



class TokenStream:
    """
    An immutable subset of a token sequence
    """

    class EndOfStream(Exception):
        pass

    Nothing = tuple()

    def __init__(self, _tokens, offset=0):
        if isinstance(_tokens, tuple):
            self._tokens = _tokens
        else:
            self._tokens = tuple(_tokens)
        self._head_index = offset

    def __len__(self):
        return len(self._tokens) - self.offset

    @property
    def head(self):
        try:
            return self._tokens[self._head_index]
        except IndexError:
            raise TokenStream.EndOfStream

    @property
    def tail(self):
        return TokenStream(self._tokens, offset=self._head_index+1)

    @property
    def offset(self):
        return self._head_index

    @property
    def at_end(self):
        return self.offset >= len(self._tokens)



from poetry.utils.venv import Venv

from .command import Command


class VenvCommand(Command):

    def __init__(self, name=None):
        self._venv = None

        super().__init__(name)

    def initialize(self, i, o):
        super().initialize(i, o)

        self._venv = Venv.create(o, self.poetry.package.name)

        if self._venv.is_venv() and o.is_verbose():
            o.writeln(f'Using virtualenv: <comment>{self._venv.venv}</>')

    @property
    def venv(self):
        return self._venv



from .helpers import flat_map


class ResolverError(Exception):

    pass


class NoSuchDependencyError(ResolverError):

    def __init__(self, dependency, required_by=None):
        if required_by is None:
            required_by = []

        sources = ' and '.join(['"{}"'.format(r) for r in required_by])
        message = 'Unable to find a specification for "{}"'.format(dependency)
        if sources:
            message += ' depended upon by {}'.format(sources)

        super().__init__(message)


class CircularDependencyError(ResolverError):

    def __init__(self, vertices):
        super(CircularDependencyError, self).__init__(
            'There is a circular dependency between {}'.format(
                ' and '.join([v.name for v in vertices])
            )
        )

        self._dependencies = [v.payload.possibilities[-1] for v in vertices]

    @property
    def dependencies(self):
        return self._dependencies


class VersionConflict(ResolverError):

    def __init__(self, conflicts, specification_provider):
        pairs = []

        for conflicting in flat_map(
            list(conflicts.values()), lambda x: x.requirements
        ):
            for source, conflict_requirements in conflicting.items():
                for c in conflict_requirements:
                    pairs.append((c, source))

        super().__init__(
            'Unable to satisfy the following requirements:\n\n'
            '{}'.format(
                '\n'.join('- "{}" required by "{}"'.format(r, d)
                          for r, d in pairs)
            )
        )

        self._conflicts = conflicts
        self._specification_provider = specification_provider

    @property
    def conflicts(self):
        return self._conflicts

    @property
    def specification_provider(self):
        return self._specification_provider

    def message_with_trees(self,
                           solver_name='Poetry',
                           possibility_type='possibility named',
                           reduce_trees=lambda trees: sorted(set(trees), key=str),
                           printable_requirement=str,
                           message_for_conflict=None,
                           version_for_spec=str):
        o = []
        for name, conflict in sorted(self._conflicts):
            o.append(
                '\n{} could not find compatible versions for {} "{}"_n'.format(
                    solver_name, possibility_type, name
                )
            )

            if conflict.locked_requirement:
                o.append(
                    ' In snapshot ({}):\n'.format(
                        self._specification_provider.name_for_locking_dependency_source
                    )
                )
                o.append(
                    '    {}\n'.format(
                        printable_requirement(conflict.locked_requirement)
                    )
                )
                o.append('\n')

            o.append(
                '  In {}:\n'.format(
                    self._specification_provider.name_for_explicit_dependency_source
                )
            )
            trees = reduce_trees(conflict.requirement_trees)
            ot = []
            for tree in trees:
                t = ''
                depth = 2
                for req in tree:
                    t += '  ' * depth + str(req)

                    if tree[-1] != req:
                        spec = conflict.activated_by_name.get(
                            self._specification_provider.name_for(req)
                        )
                        if spec:
                            t += ' was resolved to {}, which'.format(
                                version_for_spec(spec)
                            )

                        t += ' depends on'

                    t += '\n'
                    depth += 1

                ot.append(t)

            o.append('\n'.join(ot))

            if message_for_conflict:
                message_for_conflict(o, name, conflict)

        return ''.join(o).strip()





import datetime
import six

from .. import tokens
from ..tokens import py2toml
from ..util import join_with
from .atomic import AtomicElement
from .metadata import PunctuationElement, WhitespaceElement, NewlineElement
from .tableheader import TableHeaderElement


def create_element(value, multiline_strings_allowed=True):
    """
    Creates and returns the appropriate elements.Element instance from the given Python primitive, sequence-like,
    or dict-like value.
    """
    from .array import ArrayElement

    if isinstance(value, (int, float, bool, datetime.datetime, datetime.date) + six.string_types) or value is None:
        primitive_token = py2toml.create_primitive_token(value, multiline_strings_allowed=multiline_strings_allowed)
        return AtomicElement((primitive_token,))

    elif isinstance(value, (list, tuple)):
        preamble = [create_operator_element('[')]
        postable = [create_operator_element(']')]
        stuffing_elements = [create_element(v) for v in value]
        spaced_stuffing = join_with(stuffing_elements,
                                    separator=[create_operator_element(','), create_whitespace_element()])

        return ArrayElement(preamble + spaced_stuffing + postable)

    elif isinstance(value, dict):
        return create_inline_table(value, multiline_table=False, multiline_strings_allowed=multiline_strings_allowed)

    else:
        raise RuntimeError('Value type unaccounted for: {} of type {}'.format(value, type(value)))


def create_inline_table(from_dict, multiline_table=False, multiline_strings_allowed=True):
    """
    Creates an InlineTable element from the given dict instance.
    """

    from .inlinetable import InlineTableElement

    preamble = [create_operator_element('{')]
    postable = [create_operator_element('}')]

    stuffing_elements = (
        (
            create_string_element(k, bare_allowed=True),
            create_whitespace_element(),
            create_operator_element('='),
            create_whitespace_element(),
            create_element(v, multiline_strings_allowed=False)
        ) for (k, v) in from_dict.items())

    pair_separator = [create_operator_element(','),
                      create_newline_element() if multiline_table else create_whitespace_element()]
    spaced_elements = join_with(stuffing_elements, separator=pair_separator)

    return InlineTableElement(preamble + spaced_elements + postable)


def create_string_element(value, bare_allowed=False):
    """
    Creates and returns an AtomicElement wrapping a string value.
    """
    return AtomicElement((py2toml.create_string_token(value, bare_allowed),))


def create_operator_element(operator):
    """
    Creates a PunctuationElement instance containing an operator token of the specified type. The operator
    should be a TOML source str.
    """
    operator_type_map = {
        ',': tokens.TYPE_OP_COMMA,
        '=': tokens.TYPE_OP_ASSIGNMENT,
        '[': tokens.TYPE_OP_SQUARE_LEFT_BRACKET,
        ']': tokens.TYPE_OP_SQUARE_RIGHT_BRACKET,
        '[[': tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET,
        ']]': tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET,
        '{': tokens.TYPE_OP_CURLY_LEFT_BRACKET,
        '}': tokens.TYPE_OP_CURLY_RIGHT_BRACKET,
    }

    ts = (tokens.Token(operator_type_map[operator], operator),)
    return PunctuationElement(ts)


def create_newline_element():
    """
    Creates and returns a single NewlineElement.
    """
    ts = (tokens.Token(tokens.TYPE_NEWLINE, '\n'),)
    return NewlineElement(ts)


def create_whitespace_element(length=1, char=' '):
    """
    Creates and returns a WhitespaceElement containing spaces.
    """
    ts = (tokens.Token(tokens.TYPE_WHITESPACE, char),) * length
    return WhitespaceElement(ts)


def create_table_header_element(names):

    name_elements = []

    if isinstance(names, six.string_types):
        name_elements = [py2toml.create_string_token(names, bare_string_allowed=True)]
    else:
        for (i, name) in enumerate(names):
            name_elements.append(py2toml.create_string_token(name, bare_string_allowed=True))
            if i < (len(names)-1):
                name_elements.append(py2toml.operator_token(tokens.TYPE_OPT_DOT))

    return TableHeaderElement(
        [py2toml.operator_token(tokens.TYPE_OP_SQUARE_LEFT_BRACKET)] + name_elements +
        [py2toml.operator_token(tokens.TYPE_OP_SQUARE_RIGHT_BRACKET), py2toml.operator_token(tokens.TYPE_NEWLINE)],
    )


def create_array_of_tables_header_element(name):
    return TableHeaderElement((
        py2toml.operator_token(tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET),
        py2toml.create_string_token(name, bare_string_allowed=True),
        py2toml.operator_token(tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET),
        py2toml.operator_token(tokens.TYPE_NEWLINE),
    ))


def create_table(dict_value):
    """
    Creates a TableElement out of a dict instance.
    """
    from .table import TableElement

    if not isinstance(dict_value, dict):
        raise ValueError('input must be a dict instance.')

    table_element = TableElement([create_newline_element()])
    for k, v in dict_value.items():
        table_element[k] = create_element(v)

    return table_element


def create_multiline_string(text, maximum_line_length):
    return AtomicElement(_tokens=[py2toml.create_multiline_string(text, maximum_line_length)])



from .dependency import Dependency


class VCSDependency(Dependency):
    """
    Represents a VCS dependency
    """

    def __init__(self, name, vcs, source,
                 branch=None, tag=None, rev=None,
                 optional=False):
        self._vcs = vcs
        self._source = source

        if not any([branch, tag, rev]):
            # If nothing has been specified, we assume master
            branch = 'master'

        self._branch = branch
        self._tag = tag
        self._rev = rev
        
        super().__init__(
            name, '*', optional=optional, allows_prereleases=True
        )

    @property
    def vcs(self) -> str:
        return self._vcs

    @property
    def source(self):
        return self._source

    @property
    def branch(self):
        return self._branch

    @property
    def tag(self):
        return self._tag

    @property
    def rev(self):
        return self._rev

    @property
    def reference(self) -> str:
        return self._branch or self._tag or self._rev

    @property
    def pretty_constraint(self) -> str:
        if self._branch:
            what = 'branch'
            version = self._branch
        elif self._tag:
            what = 'tag'
            version = self._tag
        else:
            what = 'rev'
            version = self._rev

        return f'{what} {version}'

    def is_vcs(self) -> bool:
        return True

    def accepts_prereleases(self):
        return True



import re
import subprocess


class GitConfig:

    def __init__(self):
        config_list = subprocess.check_output(
            ['git', 'config', '-l'],
            stderr=subprocess.STDOUT
        ).decode()

        self._config = {}

        m = re.findall('(?ms)^([^=]+)=(.*?)$', config_list)
        if m:
            for group in m:
                self._config[group[0]] = group[1]

    def get(self, key, default=None):
        return self._config.get(key, default)

    def __getitem__(self, item):
        return self._config[item]


class Git:

    def __init__(self, work_dir=None):
        self._config = GitConfig()
        self._work_dir = work_dir

    @property
    def config(self) -> GitConfig:
        return self._config

    def clone(self, repository, dest) -> str:
        return self.run('clone', repository, dest)

    def checkout(self, rev, folder=None) -> str:
        args = []
        if folder is None and self._work_dir:
            folder = self._work_dir

        if folder:
            args += [
                '--git-dir', (folder / '.git').as_posix(),
                '--work-tree', folder.as_posix()
            ]

        args += [
            'checkout', rev
        ]

        return self.run(*args)

    def rev_parse(self, rev, folder=None) -> str:
        args = []
        if folder is None and self._work_dir:
            folder = self._work_dir

        if folder:
            args += [
                '--git-dir', (folder / '.git').as_posix(),
                '--work-tree', folder.as_posix()
            ]

        args += [
            'rev-parse', rev
        ]

        return self.run(*args)

    def get_ignored_files(self, folder=None) -> list:
        args = []
        if folder is None and self._work_dir:
            folder = self._work_dir

        if folder:
            args += [
                '--git-dir', (folder / '.git').as_posix(),
                '--work-tree', folder.as_posix()
            ]

        args += [
            'ls-files', '--others', '-i', '--exclude-standard'
        ]
        output = self.run(*args)

        return output.split('\n')

    def run(self, *args) -> str:
        return subprocess.check_output(
            ['git'] + list(args),
            stderr=subprocess.STDOUT
        ).decode()



from pathlib import Path

from .git import Git


def get_vcs(directory: Path):
    directory = directory.resolve()

    for p in [directory] + list(directory.parents):
        if (p / '.git').is_dir():
            return Git(p)



import json

from pathlib import Path

import jsonschema

from .__version__ import __version__
from .exceptions import InvalidProjectFile
from .packages import Dependency
from .packages import Locker
from .packages import Package
from .repositories import Pool
from .repositories.pypi_repository import PyPiRepository
from .utils.toml_file import TomlFile


class Poetry:

    VERSION = __version__

    def __init__(self,
                 file: Path,
                 config: dict,
                 package: Package,
                 locker: Locker):
        self._file = TomlFile(file)
        self._package = package
        self._config = config
        self._locker = locker

        # Configure sources
        self._pool = Pool()
        for source in self._config.get('source', []):
            self._pool.configure(source)

        # Always put PyPI last to prefere private repositories
        self._pool.add_repository(PyPiRepository())
        
    @property
    def file(self):
        return self._file

    @property
    def package(self) -> Package:
        return self._package

    @property
    def config(self) -> dict:
        return self._config

    @property
    def locker(self) -> Locker:
        return self._locker

    @property
    def pool(self) -> Pool:
        return self._pool

    @classmethod
    def create(cls, cwd) -> 'Poetry':
        poetry_file = Path(cwd) / 'pyproject.toml'

        if not poetry_file.exists():
            raise RuntimeError(
                f'Poetry could not find a pyproject.toml file in {cwd}'
            )

        local_config = TomlFile(poetry_file.as_posix()).read(True)
        if 'tool' not in local_config or 'poetry' not in local_config['tool']:
            raise RuntimeError(
                f'[tool.poetry] section not found in {poetry_file.name}'
            )
        local_config = local_config['tool']['poetry']

        # Checking validity
        cls.check(local_config)

        # Load package
        name = local_config['name']
        version = local_config['version']
        package = Package(name, version, version)

        for author in local_config['authors']:
            package.authors.append(author)

        package.description = local_config.get('description', '')
        package.homepage = local_config.get('homepage')
        package.repository_url = local_config.get('repository')
        package.license = local_config.get('license')
        package.keywords = local_config.get('keywords', [])

        if 'readme' in local_config:
            package.readme = Path(cwd) / local_config['readme']

        if 'platform' in local_config:
            package.platform = local_config['platform']

        if 'dependencies' in local_config:
            for name, constraint in local_config['dependencies'].items():
                if name.lower() == 'python':
                    package.python_versions = constraint
                    continue

                package.add_dependency(name, constraint)

        if 'dev-dependencies' in local_config:
            for name, constraint in local_config['dev-dependencies'].items():
                package.add_dependency(name, constraint, category='dev')

        if 'extras' in local_config:
            for extra_name, requirements in local_config['extras'].items():
                package.extras[extra_name] = [
                    Dependency(req, '*') for req in requirements
                ]

        if 'build' in local_config:
            package.build = local_config['build']

        if 'include' in local_config:
            package.include = local_config['include']

        if 'exclude' in local_config:
            package.exclude = local_config['exclude']

        locker = Locker(poetry_file.with_suffix('.lock'), local_config)

        return cls(poetry_file, local_config, package, locker)

    @classmethod
    def check(cls, config: dict, strict: bool = False):
        """
        Checks the validity of a configuration
        """
        schema = (
            Path(__file__).parent
            / 'json' / 'schemas' / 'poetry-schema.json'
        )

        schema = json.loads(schema.read_text())

        try:
            jsonschema.validate(
                config,
                schema
            )
        except jsonschema.ValidationError as e:
            message = e.message
            if e.path:
                message = f"[{'.'.join(e.path)}] {message}"

            raise InvalidProjectFile(message)

        return True



from poetry.installation import Installer

from .venv_command import VenvCommand


class UpdateCommand(VenvCommand):
    """
    Update dependencies as according to the <comment>pyproject.toml</> file.

    update
        { packages?* : The packages to update. }
        { --no-dev : Do not install dev dependencies. }
        { --dry-run : Outputs the operations but will not execute anything
                      (implicitly enables --verbose). }
    """

    def handle(self):
        packages = self.argument('packages')

        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool
        )

        if packages:
            installer.whitelist({name: '*' for name in packages})

        installer.dev_mode(not self.option('no-dev'))
        installer.dry_run(self.option('dry-run'))

        # Force update
        installer.update(True)

        return installer.run()



from typing import List
from typing import Union

import poetry.packages

from .base_repository import BaseRepository
from .repository import Repository


class Pool(BaseRepository):

    def __init__(self, repositories: Union[list, None] = None):
        if repositories is None:
            repositories = []

        self._repositories = []

        for repository in repositories:
            self.add_repository(repository)

        super().__init__()
            
    @property
    def repositories(self) -> List[Repository]:
        return self._repositories

    def add_repository(self, repository: Repository) -> 'Pool':
        """
        Adds a repository to the pool.
        """
        self._repositories.append(repository)

        return self
    
    def configure(self, source: dict) -> 'Pool':
        """
        Configures a repository based on a source
        specification and add it to the pool.
        """
        from .legacy_repository import LegacyRepository

        if 'url' in source:
            # PyPI-like repository
            if 'name' not in source:
                raise RuntimeError('Missing [name] in source.')

            repository = LegacyRepository(source['name'], source['url'])
        else:
            raise RuntimeError('Unsupported source specified')

        return self.add_repository(repository)

    def has_package(self, package):
        raise NotImplementedError()

    def package(self, name, version) -> Union['poetry.packages.Package', None]:
        package = poetry.packages.Package(name, version, version)
        if package in self._packages:
            return self._packages[self._packages.index(package)]

        for repository in self._repositories:
            package = repository.package(name, version)
            if package:
                self._packages.append(package)

                return package

        return None

    def find_packages(self,
                      name,
                      constraint=None,
                      extras=None) -> List['poetry.packages.Package']:
        for repository in self._repositories:
            packages = repository.find_packages(name, constraint, extras=extras)
            if packages:
                return packages

        return []

    def search(self, query, mode=BaseRepository.SEARCH_FULLTEXT):
        raise NotImplementedError()



from collections import namedtuple


class UnwindDetails:

    def __init__(self,
                 state_index,
                 state_requirement,
                 requirement_tree,
                 conflicting_requirements,
                 requirement_trees,
                 requirements_unwound_to_instead):
        self.state_index = state_index
        self.state_requirement = state_requirement
        self.requirement_tree = requirement_tree
        self.conflicting_requirements = conflicting_requirements
        self.requirement_trees = requirement_trees
        self.requirements_unwound_to_instead = requirements_unwound_to_instead
        self._reversed_requirement_tree_index = None
        self._sub_dependencies_to_avoid = None
        self._all_requirements = None

    @property
    def reversed_requirement_tree_index(self):
        if self._reversed_requirement_tree_index is None:
            if self.state_requirement:
                self._reversed_requirement_tree_index = list(reversed(
                    self.requirement_tree
                )).index(self.state_requirement)
            else:
                self._reversed_requirement_tree_index = 999999

        return self._reversed_requirement_tree_index

    def unwinding_to_primary_requirement(self):
        return self.requirement_tree[-1] == self.state_requirement

    @property
    def sub_dependencies_to_avoid(self):
        if self._sub_dependencies_to_avoid is None:
            self._sub_dependencies_to_avoid = []
            for tree in self.requirement_trees:
                try:
                    index = tree.index(self.state_requirement)
                except ValueError:
                    continue

                if tree[index + 1] is not None:
                    self._sub_dependencies_to_avoid.append(tree[index + 1])

        return self._sub_dependencies_to_avoid
    
    @property
    def all_requirements(self):
        if self._all_requirements is None:
            self._all_requirements = [
                x
                for tree in self.requirement_trees
                for x in tree
            ]

        return self._all_requirements

    def __eq__(self, other):
        if not isinstance(other, UnwindDetails):
            return NotImplemented

        return (
            self.state_index == other.state_index
            and (
                self.reversed_requirement_tree_index
                == other.reversed_requirement_tree_index
            )
        )

    def __lt__(self, other):
        if not isinstance(other, UnwindDetails):
            return NotImplemented

        return self.state_index < other.state_index

    def __le__(self, other):
        if not isinstance(other, UnwindDetails):
            return NotImplemented

        return self.state_index <= other.state_index

    def __gt__(self, other):
        if not isinstance(other, UnwindDetails):
            return NotImplemented

        return self.state_index > other.state_index

    def __ge__(self, other):
        if not isinstance(other, UnwindDetails):
            return NotImplemented

        return self.state_index >= other.state_index

    def __hash__(self):
        return hash((id(self), self.state_index, self.state_requirement))



from .prettify.errors import NoArrayFoundError
from . import structurer, toplevels, raw
from .array import ArrayOfTables
from .freshtable import FreshTable

from .prettify.elements import factory as element_factory
from .prettify import util


class TOMLFile(dict):
    """
    A TOMLFile object that tries its best to prserve formatting and order of mappings of the input source.

    Raises InvalidTOMLFileError on invalid input elements.

    Raises DuplicateKeysError, DuplicateTableError when appropriate.
    """

    def __init__(self, _elements):
        self._elements = []
        self._navigable = {}
        self.append_elements(_elements)

    def __getitem__(self, item):
        try:
            value = self._navigable[item]
            if isinstance(value, (list, tuple)):
                return ArrayOfTables(toml_file=self, name=item, iterable=value)
            else:
                return value
        except KeyError:
            return FreshTable(parent=self, name=item, is_array=False)

    def __contains__(self, item):
        return item in self.keys()

    def _setitem_with_key_seq(self, key_seq, value):
        """
        Sets a the value in the TOML file located by the given key sequence.

        Example:
        self._setitem(('key1', 'key2', 'key3'), 'text_value')
        is equivalent to doing
        self['key1']['key2']['key3'] = 'text_value'
        """
        table = self
        key_so_far = tuple()
        for key in key_seq[:-1]:
            key_so_far += (key,)
            self._make_sure_table_exists(key_so_far)
            table = table[key]
        table[key_seq[-1]] = value

    def _array_setitem_with_key_seq(self, array_name, index, key_seq, value):
        """
        Sets a the array value in the TOML file located by the given key sequence.

        Example:
        self._array_setitem(array_name, index, ('key1', 'key2', 'key3'), 'text_value')
        is equivalent to doing
        self.array(array_name)[index]['key1']['key2']['key3'] = 'text_value'
        """
        table = self.array(array_name)[index]
        key_so_far = tuple()
        for key in key_seq[:-1]:
            key_so_far += (key,)
            new_table = self._array_make_sure_table_exists(array_name, index, key_so_far)
            if new_table is not None:
                table = new_table
            else:
                table = table[key]
        table[key_seq[-1]] = value

    def _make_sure_table_exists(self, name_seq):
        """
        Makes sure the table with the full name comprising of name_seq exists.
        """
        t = self
        for key in name_seq[:-1]:
            t = t[key]
        name = name_seq[-1]
        if name not in t:
            self.append_elements([element_factory.create_table_header_element(name_seq),
                                  element_factory.create_table({})])

    def _array_make_sure_table_exists(self, array_name, index, name_seq):
        """
        Makes sure the table with the full name comprising of name_seq exists.
        """
        t = self[array_name][index]
        for key in name_seq[:-1]:
            t = t[key]
        name = name_seq[-1]
        if name not in t:
            new_table = element_factory.create_table({})
            self.append_elements([element_factory.create_table_header_element((array_name,) + name_seq), new_table])
            return new_table

    def __delitem__(self, key):
        table_element_index = self._elements.index(self._navigable[key])
        self._elements[table_element_index] = element_factory.create_table({})
        self._on_element_change()

    def __setitem__(self, key, value):

        # Setting an array-of-tables
        if key and isinstance(value, (tuple, list)) and value and all(isinstance(v, dict) for v in value):
            for table in value:
                self.array(key).append(table)

        # Or setting a whole single table
        elif isinstance(value, dict):

            if key and key in self:
                del self[key]

            for key_seq, child_value in util.flatten_nested({key: value}).items():
                self._setitem_with_key_seq(key_seq, child_value)

            # if key in self._navigable:
            #     del self[key]
            #     index = self._elements.index(self._navigable[key])
            #     self._elements = self._elements[:index] + [element_factory.create_table(value)] + self._elements[index+1:]
            # else:
            #     if key:
            #         self._elements.append(element_factory.create_table_header_element(key))
            #     self._elements.append(element_factory.create_table(value))


        # Or updating the anonymous section table
        else:
            # It's mea
            self[''][key] = value

        self._on_element_change()

    def _detect_toplevels(self):
        """
        Returns a sequence of TopLevel instances for the current state of this table.
        """
        return tuple(e for e in toplevels.identify(self.elements) if isinstance(e, toplevels.Table))

    def _update_table_fallbacks(self, table_toplevels):
        """
        Updates the fallbacks on all the table elements to make relative table access possible.

        Raises DuplicateKeysError if appropriate.
        """

        if len(self.elements) <= 1:
            return

        def parent_of(toplevel):
            # Returns an TopLevel parent of the given entry, or None.
            for parent_toplevel in table_toplevels:
                if toplevel.name.sub_names[:-1] == parent_toplevel.name.sub_names:
                    return parent_toplevel

        for entry in table_toplevels:
            if entry.name.is_qualified:
                parent = parent_of(entry)
                if parent:
                    child_name = entry.name.without_prefix(parent.name)
                    parent.table_element.set_fallback({child_name.sub_names[0]: entry.table_element})

    def _recreate_navigable(self):
        if self._elements:
            self._navigable = structurer.structure(toplevels.identify(self._elements))

    def array(self, name):
        """
        Returns the array of tables with the given name.
        """
        if name in self._navigable:
            if isinstance(self._navigable[name], (list, tuple)):
                return self[name]
            else:
                raise NoArrayFoundError
        else:
            return ArrayOfTables(toml_file=self, name=name)

    def _on_element_change(self):
        self._recreate_navigable()

        table_toplevels = self._detect_toplevels()
        self._update_table_fallbacks(table_toplevels)

    def append_elements(self, elements):
        """
        Appends more elements to the contained internal elements.
        """
        self._elements = self._elements + list(elements)
        self._on_element_change()

    def prepend_elements(self, elements):
        """
        Prepends more elements to the contained internal elements.
        """
        self._elements = list(elements) + self._elements
        self._on_element_change()

    def dumps(self):
        """
        Returns the TOML file serialized back to str.
        """
        return ''.join(element.serialized() for element in self._elements)

    def dump(self, file_path):
        with open(file_path, mode='w') as fp:
            fp.write(self.dumps())

    def keys(self):
        return set(self._navigable.keys()) | {''}

    def values(self):
        return self._navigable.values()

    def items(self):
        items = list(self._navigable.items())

        def has_anonymous_entry():
            return any(key == '' for (key, _) in items)

        if has_anonymous_entry():
            return items
        else:
            return items + [('', self[''])]

    def get(self, item, default=None):
        return self._navigable.get(item, default)

    @property
    def primitive(self):
        """
        Returns a primitive object representation for this container (which is a dict).

        WARNING: The returned container does not contain any markup or formatting metadata.
        """
        raw_container = raw.to_raw(self._navigable)

        # Collapsing the anonymous table onto the top-level container is present
        if '' in raw_container:
            raw_container.update(raw_container[''])
            del raw_container['']

        return raw_container

    def append_fresh_table(self, fresh_table):
        """
        Gets called by FreshTable instances when they get written to.
        """
        if fresh_table.name:
            elements = []
            if fresh_table.is_array:
                elements += [element_factory.create_array_of_tables_header_element(fresh_table.name)]
            else:
                elements += [element_factory.create_table_header_element(fresh_table.name)]

            elements += [fresh_table, element_factory.create_newline_element()]
            self.append_elements(elements)

        else:
            # It's an anonymous table
            self.prepend_elements([fresh_table, element_factory.create_newline_element()])

    @property
    def elements(self):
        return self._elements

    def __str__(self):

        is_empty = (not self['']) and (not tuple(k for k in self.keys() if k))

        def key_name(key):
            return '[ANONYMOUS]' if not key else key

        def pair(key, value):
            return '%s = %s' % (key_name(key), str(value))

        content_text = '' if is_empty else \
            '\n\t' + ',\n\t'.join(pair(k, v) for (k, v) in self.items() if v) + '\n'

        return "TOMLFile{%s}" % content_text

    def __repr__(self):
        return str(self)






from poetry.semver import statisfies
from poetry.version.version_selector import VersionSelector

from .venv_command import VenvCommand


class ShowCommand(VenvCommand):
    """
    Shows information about packages.

    show
        { package? : Package to inspect. }
        { --t|tree : List the dependencies as a tree. }
        { --l|latest : Show the latest version. }
        { --o|outdated : Show the latest version
                         but only for packages that are outdated. }
    """

    help = """The show command displays detailed information about a package, or
lists all packages available."""

    colors = [
        'green',
        'yellow',
        'cyan',
        'magenta',
        'blue',
    ]

    def handle(self):
        package = self.argument('package')

        if self.option('tree'):
            self.init_styles()

        if self.option('outdated'):
            self.input.set_option('latest', True)

        installed_repo = self.poetry.locker.locked_repository(True)

        # Show tree view if requested
        if self.option('tree') and not package:
            requires = self.poetry.package.requires + self.poetry.package.dev_requires
            packages = installed_repo.packages
            for package in packages:
                for require in requires:
                    if package.name == require.name:
                        self.display_package_tree(package, installed_repo)
                        break

            return 0

        table = self.table(style='compact')
        table.get_style().set_vertical_border_char('')
        locked_packages = installed_repo.packages

        if package:
            pkg = None
            for locked in locked_packages:
                if package.lower() == locked.name:
                    pkg = locked
                    break

            if not pkg:
                raise ValueError(f'Package {package} not found')

            if self.option('tree'):
                self.display_package_tree(pkg, installed_repo)

                return 0

            rows = [
                ['<info>name</>', f' : <fg=cyan>{pkg.pretty_name}</>'],
                ['<info>version</>', f' : <comment>{pkg.pretty_version}</>'],
                ['<info>description</>', f' : {pkg.description}'],
            ]

            table.add_rows(rows)
            table.render()

            if pkg.requires:
                self.line('')
                self.line('<info>dependencies</info>')
                for dependency in pkg.requires:
                    self.line(f' - {dependency.pretty_name} '
                              f'<comment>{dependency.pretty_constraint}</>')

            return 0

        show_latest = self.option('latest')
        terminal = self.get_application().terminal
        width = terminal.width
        name_length = version_length = latest_length = 0
        latest_packages = {}
        # Computing widths
        for locked in locked_packages:
            name_length = max(name_length, len(locked.pretty_name))
            version_length = max(version_length, len(locked.full_pretty_version))
            if show_latest:
                latest = self.find_latest_package(locked)
                if not latest:
                    latest = locked

                latest_packages[locked.pretty_name] = latest
                latest_length = max(latest_length, len(latest.full_pretty_version))

        write_version = name_length + version_length + 3 <= width
        write_latest = name_length + version_length + latest_length + 3 <= width
        write_description = name_length + version_length + latest_length + 24 <= width

        for locked in locked_packages:
            line = f'<fg=cyan>{locked.pretty_name:{name_length}}</>'
            if write_version:
                line += f' {locked.full_pretty_version:{version_length}}'
            if show_latest and write_latest:
                latest = latest_packages[locked.pretty_name]

                update_status = self.get_update_status(latest, locked)
                color = 'green'
                if update_status == 'semver-safe-update':
                    color = 'red'
                elif update_status == 'update-possible':
                    color = 'yellow'

                line += f' <fg={color}>{latest.version:{latest_length}}</>'
                if self.option('outdated') and update_status == 'up-to-date':
                    continue

            if write_description:
                description = locked.description
                remaining = width - name_length - version_length - 4
                if show_latest:
                    remaining -= latest_length

                if len(locked.description) > remaining:
                    description = description[:remaining-3] + '...'

                line += ' ' + description

            self.line(line)

    def display_package_tree(self, package, installed_repo):
        self.write(f'<info>{package.pretty_name}</info>')
        self.line(f' {package.pretty_version} {package.description}')

        dependencies = package.requires
        dependencies = sorted(dependencies, key=lambda x: x.name)
        tree_bar = '├'
        j = 0
        total = len(dependencies)
        for dependency in dependencies:
            j += 1
            if j == total:
                tree_bar = '└'

            level = 1
            color = self.colors[level]
            info = f'{tree_bar}── <{color}>{dependency.name}</{color}> ' \
                   f'{dependency.pretty_constraint}'
            self._write_tree_line(info)

            tree_bar = tree_bar.replace('└', ' ')
            packages_in_tree = [package.name, dependency.name]

            self._display_tree(
                dependency, installed_repo, packages_in_tree,
                tree_bar, level + 1
            )

    def _display_tree(self,
                      dependency, installed_repo, packages_in_tree,
                      previous_tree_bar='├', level=1):
        previous_tree_bar = previous_tree_bar.replace('├', '│')

        dependencies = []
        for package in installed_repo.packages:
            if package.name == dependency.name:
                dependencies = package.requires

                break

        dependencies = sorted(dependencies, key=lambda x: x.name)
        tree_bar = previous_tree_bar + '   ├'
        i = 0
        total = len(dependencies)
        for dependency in dependencies:
            i += 1
            current_tree = packages_in_tree
            if i == total:
                tree_bar = previous_tree_bar + '   └'

            color_ident = level % len(self.colors)
            color = self.colors[color_ident]

            circular_warn = ''
            if dependency.name in current_tree:
                circular_warn = '(circular dependency aborted here)'

            info = f'{tree_bar}── <{color}>{dependency.name}</{color}> ' \
                   f'{dependency.pretty_constraint} {circular_warn}'
            self._write_tree_line(info)

            tree_bar = tree_bar.replace('└', ' ')

            if dependency.name not in current_tree:
                current_tree.append(dependency.name)

                self._display_tree(
                    dependency, installed_repo, current_tree,
                    tree_bar, level + 1
                )

    def _write_tree_line(self, line):
        if not self.output.is_decorated():
            line = line.replace('└', '`-')
            line = line.replace('├', '|-')
            line = line.replace('──', '-')
            line = line.replace('│', '|')

        self.line(line)

    def init_styles(self):
        for color in self.colors:
            self.set_style(color, color)

    def find_latest_package(self, package):
        # find the latest version allowed in this pool
        if package.source_type == 'git':
            return

        name = package.name
        selector = VersionSelector(self.poetry.pool)

        return selector.find_best_candidate(name, f'>={package.version}')

    def get_update_status(self, latest, package):
        if latest.full_pretty_version == package.full_pretty_version:
            return 'up-to-date'

        constraint = '^' + package.version

        if latest.version and statisfies(latest.version, constraint):
            # It needs an immediate semver-compliant upgrade
            return 'semver-safe-update'

        # it needs an upgrade but has potential BC breaks so is not urgent
        return 'update-possible'



from poetry.installation import Installer

from .venv_command import VenvCommand


class InstallCommand(VenvCommand):
    """
    Installs the project dependencies.

    install
        { --no-dev : Do not install dev dependencies. }
        { --dry-run : Outputs the operations but will not execute anything
                      (implicitly enables --verbose). }
        { --E|extras=* : Extra sets of dependencies to install
                         (multiple values allowed). }
    """

    help = """The <info>install</info> command reads the <comment>pyproject.toml</> file from
the current directory, processes it, and downloads and installs all the
libraries and dependencies outlined in that file. If the file does not
exist it will look for <comment>pyproject.toml</> and do the same.

<info>poetry install</info>    
"""

    def handle(self):
        installer = Installer(
            self.output,
            self.venv,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool
        )

        installer.extras(self.option('extras'))
        installer.dev_mode(not self.option('no-dev'))
        installer.dry_run(self.option('dry-run'))
        installer.verbose(self.option('verbose'))

        return installer.run()
